from numba import njit
from numba.typed import Dict
from numba.core import types
from numpy import empty_like, abs, nan, isfinite, flip, linspace, sqrt
from numpy.typing import NDArray
from typing import Literal
from numpy import interp
from scipy.signal import sosfilt, sosfiltfilt, butter


@njit(cache=True)
def exponential_filter(time_seconds, signal, options: dict = None):
    """
    Exponential filter that operates on the differences between successive values.

    :param time_seconds:
    :param signal:
    :param options:
    :return:
    """

    if options is None:
        options = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

    smoothing_factor = options.get("smoothing_factor", 0.004)
    maximum_gap_size = options.get("maximum_gap_size", 3)
    sampling_frequency = options.get("sampling_frequency", 2.5)
    sampling_interval = 1 / sampling_frequency

    # Initialize empty array
    filtered_signal = empty_like(signal)

    # Initialize start values
    exponential_mean = 0
    filtered_signal[0] = 0

    for ii in range(1, len(signal)):
        if (
            time_seconds[ii] - time_seconds[ii - 1]
            > maximum_gap_size * sampling_interval
        ) or (not isfinite(signal[ii])):
            # Restart
            exponential_mean = 0.0
            filtered_signal[ii] = 0.0
            continue

        exponential_mean = (
            exponential_mean * (1 - smoothing_factor) + signal[ii] * smoothing_factor
        )
        filtered_signal[ii] = signal[ii] - exponential_mean

    return filtered_signal


@njit(cache=True)
def exponential_delta_filter(time_seconds, signal, options: dict = None):
    """
    Exponential filter that operates on the differences between succesive values.

    :param time_seconds:
    :param signal:
    :param options:
    :return:
    """

    if options is None:
        options = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

    smoothing_factor = options.get("smoothing_factor", 0.004)
    maximum_gap_size = options.get("maximum_gap_size", 3)
    sampling_frequency = options.get("sampling_frequency", 2.5)
    sampling_interval = 1 / sampling_frequency

    # Initialize empty array
    filtered_signal = empty_like(signal)

    # Initialize start values
    exponential_mean = 0
    filtered_signal[0] = 0

    for ii in range(1, len(signal)):
        signal_delta = signal[ii] - signal[ii - 1]

        if (
            time_seconds[ii] - time_seconds[ii - 1]
            > maximum_gap_size * sampling_interval
        ) or (not isfinite(signal_delta)):
            # Restart
            exponential_mean = 0.0
            filtered_signal[ii] = 0.0
            continue

        exponential_mean = (
            exponential_mean * (1 - smoothing_factor) + signal_delta * smoothing_factor
        )
        filtered_signal[ii] = filtered_signal[ii - 1] + (
            signal_delta - exponential_mean
        )

    return filtered_signal


@njit(cache=True)
def spike_filter(time: NDArray, signal: NDArray, options: dict = None) -> NDArray:
    if options is None:
        options = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

    rejection_threshold = options.get("rejection_threshold", 9.81)

    filtered_signal = empty_like(signal)
    filtered_signal[0] = signal[0]

    prev_signal = signal[0]
    prev_time = time[0]

    for ii in range(1, len(signal)):
        delta_signal = signal[ii] - prev_signal
        delta_time = time[ii] - prev_time

        if abs(delta_signal / delta_time) > rejection_threshold:
            filtered_signal[ii] = nan
            continue

        filtered_signal[ii] = prev_signal + delta_signal
        prev_signal = filtered_signal[ii]
        prev_time = time[ii]
    return nan_interpolate(time, filtered_signal)


@njit(cache=True)
def cumulative_filter(signal: NDArray, options: dict = None) -> NDArray:
    """
    Filter an input signal to remove step-like changes according to a cumulative filter approach.

    :param signal: Input signal at a constant sampling interval. The signal and its first order difference are
    observations of zero-mean processes.
    :param options: optional dictionary to set algorithm parameters.
    :return: Filtered signal with step like changes to the mean removed.
    """

    if options is None:
        options = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

    smoothing_factor = options.get("cumulative_smoothing_factor", 0.01)
    scale_factor = options.get("scale_factor", 5.0)
    correct = options.get("correct", 1.0)

    # Allocate the output filtered signal and set the first value to the first entry in the signal.
    filtered_signal = empty_like(signal)
    filtered_signal[0] = signal[0]

    # Initialize algorithm paramteres
    cumulative_distance = (
        0.0  # Cumulative change in the function since the last delta-zero-crossing
    )
    prev_velocity = 0.0  # previous difference
    idx_last_zero = 0  # Index of the last sign change
    variance = 4  # Variance in the cumulative difference.
    change_in_mean_detected = False  # Did we undergo a likely stepchange in the current upward or downward drift of the
    # signal?

    for current_idx in range(1, len(signal)):
        velocity = signal[current_idx] - signal[current_idx - 1]

        if prev_velocity * velocity < 0:
            # flip in sign of the differences indicates the "velocity" changed sign. Or a zero-crossing of the velocity/
            # difference signal.

            if not change_in_mean_detected:
                # No step change. Update the mean of the variance.
                variance = (
                    variance * (1 - smoothing_factor)
                    + cumulative_distance**2 * smoothing_factor
                )

            else:
                # Step change. Remove the linear trend from the current position to the last known zero-crossing.
                if correct == 1.0:
                    filtered_signal[idx_last_zero:current_idx] = filtered_signal[
                        idx_last_zero:current_idx
                    ] - cumulative_distance * linspace(
                        0, 1, current_idx - idx_last_zero
                    )
                else:
                    filtered_signal[idx_last_zero + 1 : current_idx - 1] = nan
                    filtered_signal[current_idx - 1] = filtered_signal[idx_last_zero]

            idx_last_zero = current_idx
            cumulative_distance = 0.0
            change_in_mean_detected = False
        else:
            #
            # Check how far have we moved up (or down) since the last channge in sign of the velocity signal. If the
            # cumulative change is larger than a typical change we assume that there has been a change in the mean. Note
            # That no corrective action is taken here. We only correct the signal once velocities have flipped sign.
            if abs(cumulative_distance) > scale_factor * sqrt(variance):
                change_in_mean_detected = True

        # Update previous velocity, cumulative signal and the filtered signal.
        prev_velocity = velocity
        cumulative_distance += velocity
        filtered_signal[current_idx] = filtered_signal[current_idx - 1] + velocity

    return filtered_signal


@njit(cache=True)
def cumsum(signal: NDArray, options: dict = None) -> NDArray:
    """
    Filter an input signal to remove step-like changes according to a cumulative filter approach.

    :param signal: Input signal at a constant sampling interval. The signal and its first order difference are
    observations of zero-mean processes.
    :param options: optional dictionary to set algorithm parameters.
    :return: Filtered signal with step like changes to the mean removed.
    """

    if options is None:
        options = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

    smoothing_factor = options.get("cumulative_smoothing_factor", 0.004)
    # scale_factor = options.get("scale_factor", 5.0)
    # correct = options.get("correct", 0.0)

    # Allocate the output filtered signal and set the first value to the first entry in the signal.
    filtered_signal = empty_like(signal)
    filtered_signal[0] = signal[0]

    # Initialize algorithm paramteres
    cumsum_h = 0.0
    cumsum_l = 0.0

    idx_last_zero = 0  # Index of the last sign change

    signal_variance = 0.0
    signal_mean = 0.0
    drift = 0.0

    prev_drift = 0
    was_unstable = False
    last_stable_drift = 0
    for current_idx in range(1, len(signal)):
        velocity = signal[current_idx] - signal[current_idx - 1]

        signal_mean = signal_mean * (1 - smoothing_factor) + velocity * smoothing_factor
        signal_variance = (
            signal_variance * (1 - smoothing_factor)
            + (signal[current_idx] - signal_mean) ** 2
        )

        cumsum_h = cumsum_h + (
            velocity - signal_mean
        )  # - 0.000001  #(velocity - signal_mean)
        cumsum_l = cumsum_l - (
            velocity - signal_mean
        )  # - 0.000001   #(velocity - signal_mean)

        if cumsum_h < 0.0:
            drift += cumsum_h
            cumsum_h = 0.0

        if cumsum_l < 0.0:
            drift -= cumsum_l
            cumsum_l = 0.0

        if drift == prev_drift:

            if was_unstable:
                if abs(drift - last_stable_drift) > 2:
                    filtered_signal[idx_last_zero + 1 : current_idx - 1] = -10
                    filtered_signal[current_idx - 1] = filtered_signal[idx_last_zero]

            idx_last_zero = current_idx
            was_unstable = False
            last_stable_drift = drift
        else:
            was_unstable = True

        prev_drift = drift
        filtered_signal[current_idx] = filtered_signal[current_idx - 1] + velocity

    return filtered_signal


def sos_filter(
    signal: NDArray, direction: Literal["backward", "forward", "filtfilt"], **kwargs
) -> NDArray:
    #
    # Apply forward/backward/filtfilt sos filter
    #
    # Get SOS coefficients

    mask = ~isfinite(signal)
    output = signal.copy()
    output = nan_interpolate(
        linspace(0, len(signal), len(signal), endpoint=False), output
    )

    sos = kwargs.get("sos", None)
    if sos is None:
        sos = butter(4, 0.033, btype="high", output="sos", fs=2.5)

    if direction == "backward":
        output = flip(sosfilt(sos, flip(output)))

    elif direction == "forward":
        output = sosfilt(sos, output)

    elif direction == "filtfilt":
        output = sosfiltfilt(sos, output)

    else:
        raise ValueError(f"Unknown direction {direction}")
    output[mask] = nan
    return output


@njit(cache=True)
def nan_interpolate(time: NDArray, signal: NDArray) -> NDArray:
    mask = isfinite(signal)
    return interp(time, time[mask], signal[mask])



from typing import Tuple, Callable, List
import numpy
from roguewave.tools.math import wrapped_difference
from roguewave.tools.grid import enclosing_points_1d
from roguewave.interpolate.general import interpolation_weights_1d


class NdInterpolator:
    def __init__(
        self,
        get_data: Callable[[List[numpy.ndarray], List[int]], numpy.ndarray],
        data_coordinates,
        data_shape,
        interp_coord_names,
        interp_index_coord_name: str,
        data_periodic_coordinates,
        data_period=None,
        data_discont=None,
    ):

        self.get_data = get_data
        self.coord = [x[0] for x in data_coordinates]
        self.data_shape = data_shape
        self.interp_coord_names = interp_coord_names
        self.interp_index_coord_name = interp_index_coord_name
        self.data_coordinates = data_coordinates
        self.data_periodic_coordinates = data_periodic_coordinates
        self.data_period = data_period
        self.data_discont = data_discont

    @property
    def passive_coordinate_names(self):
        return [name for name in self.coord if name not in self.interp_coord_names]

    @property
    def passive_coord_dim_indices(self) -> List[int]:
        return [self.coord.index(x) for x in self.passive_coordinate_names]

    @property
    def output_passive_coord_dim_indices(self) -> Tuple[int]:
        indices = list(range(self.output_ndims))
        _ = indices.pop(indices.index(self.output_index_coord_index))
        return tuple(indices)

    @property
    def interp_coord_dim_indices(self) -> List[int]:
        return [self.coord.index(x) for x in self.interp_coord_names]

    @property
    def interp_index_coord_index(self):
        return self.coord.index(self.interp_index_coord_name)

    def output_shape(self, number_of_points) -> numpy.ndarray:
        output_shape = numpy.ones(self.output_ndims, dtype="int32")
        interpolating_index = self.output_index_coord_index
        passive_ind = self.passive_coord_dim_indices
        jj = 0
        for index in range(self.output_ndims):
            if index == interpolating_index:
                output_shape[index] = number_of_points
            else:
                output_shape[index] = self.data_shape[passive_ind[jj]]
                jj += 1
        return output_shape

    @property
    def output_index_coord_index(self) -> int:
        return numpy.searchsorted(
            self.passive_coord_dim_indices, self.interp_index_coord_index
        )

    @property
    def interpolating_coordinates(self) -> List[Tuple[str, numpy.ndarray]]:
        return [x for x in self.data_coordinates if x[0] in self.interp_coord_names]

    @property
    def output_ndims(self):
        return self.data_ndims - self.interp_ndims + 1

    @property
    def interp_ndims(self):
        return len(self.interp_coord_names)

    @property
    def data_ndims(self):
        return len(self.coord)

    def output_indexing_full(self, slicer):
        indicer = [slice(None)] * self.output_ndims
        indicer[self.output_index_coord_index] = slicer
        return tuple(indicer)

    def output_indexing_broadcast(self, slicer):
        indicer = [None] * self.output_ndims
        indicer[self.output_index_coord_index] = slicer
        return tuple(indicer)

    def coordinate_period(self, coordinate_name):
        if coordinate_name in self.data_periodic_coordinates:
            return self.data_periodic_coordinates[coordinate_name]
        else:
            return None

    @property
    def data_is_periodic(self):
        return self.data_period is not None

    def interpolate(
        self,
        points,
    ):
        """

        :param self:
        :param interpolatinc_loc:
        :param periodic_coordinates:
        :param period_data:
        :param discont:
        :return:
        """
        number_points = len(points[self.interp_coord_names[0]])

        # Find indices and weights for the succesive 1d interpolation problems
        indices_1d = numpy.empty((self.interp_ndims, 2, number_points), dtype="int64")

        weights_1d = numpy.empty((self.interp_ndims, 2, number_points), dtype="float64")

        for index, (coordinate_name, coordinate) in enumerate(
            self.interpolating_coordinates
        ):

            period = self.coordinate_period(coordinate_name)
            indices_1d[index, :, :] = enclosing_points_1d(
                coordinate, points[coordinate_name], period=period
            )
            weights_1d[index, :, :] = interpolation_weights_1d(
                coordinate,
                points[coordinate_name],
                indices_1d[index, :, :],
                period=period,
                extrapolate_left=False,
                extrapolate_right=False,
            )

        if self.data_is_periodic:
            return self._periodic_data_interpolator(
                number_points, indices_1d, weights_1d
            )
        else:
            return self._data_interpolator(number_points, indices_1d, weights_1d)

    def _data_interpolator(self, number_points, indices_1d, weights_1d):

        # We keep a running sum of the weights, if a point is excluded because it
        # contains no data (NaN) the weights will no longer add up to 1 - and we
        # reschale to account for the missing value. This is an easy way to account
        # for interpolation near missing points. Note that if the contribution of
        # missing weights ( 1-weights_sum) exceeds 0.5 - we consider the point
        # invalid.
        output_shape = self.output_shape(number_points)
        weights_sum = numpy.zeros(output_shape)
        interp_val = numpy.zeros(output_shape, dtype=numpy.float64)

        for intp_indices_nd, intp_weight_nd in _next_point(
            self.interp_ndims, indices_1d, weights_1d
        ):
            # Loop over all interpolation points one at a time.
            val = self.get_data(intp_indices_nd, self.interp_coord_dim_indices)

            mask = numpy.all(
                ~numpy.isnan(val), axis=self.output_passive_coord_dim_indices
            ) & (intp_weight_nd > 0)

            weights_sum[self.output_indexing_full(mask)] += intp_weight_nd[
                self.output_indexing_broadcast(mask)
            ]

            interp_val[self.output_indexing_full(mask)] += (
                intp_weight_nd[self.output_indexing_broadcast(mask)]
                * val[self.output_indexing_full(mask)]
            )

        with numpy.errstate(invalid="ignore", divide="ignore"):
            return numpy.where(weights_sum > 0.5, interp_val / weights_sum, numpy.nan)

    def _periodic_data_interpolator(self, number_points, indices_1d, weights_1d):
        # We keep a running sum of the weights, if a point is excluded because it
        # contains no data (NaN) the weights will no longer add up to 1 - and we
        # reschale to account for the missing value. This is an easy way to account
        # for interpolation near missing points. Note that if the contribution of
        # missing weights ( 1-weights_sum) exceeds 0.5 - we consider the point
        # invalid.
        output_shape = self.output_shape(number_points)
        weights_sum = numpy.zeros(output_shape)

        interp_val = numpy.zeros(output_shape, dtype=numpy.complex64)
        for intp_indices_nd, intp_weight_nd in _next_point(
            self.interp_ndims, indices_1d, weights_1d
        ):
            # Loop over all interpolation points one at a time.
            to_rad = numpy.pi * 2 / self.data_period
            val = numpy.exp(
                1j
                * self.get_data(intp_indices_nd, self.interp_coord_dim_indices)
                * to_rad
            )

            mask = numpy.all(
                ~numpy.isnan(val), axis=self.output_passive_coord_dim_indices
            )

            weights_sum[self.output_indexing_full(mask)] += intp_weight_nd[
                self.output_indexing_broadcast(mask)
            ]

            interp_val[self.output_indexing_full(mask)] += (
                intp_weight_nd[self.output_indexing_broadcast(mask)]
                * val[self.output_indexing_full(mask)]
            )

        interp_val = (
            numpy.angle(
                numpy.where(weights_sum > 0.5, interp_val / weights_sum, numpy.nan)
            )
            * self.data_period
            / numpy.pi
            / 2
        )

        return wrapped_difference(
            delta=interp_val, period=self.data_period, discont=self.data_period
        )


def _next_point(
    recursion_depth, indices_1d: numpy.ndarray, weights_1d: numpy.ndarray, *narg
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    We are trying to interpolate over N dimensions. In bilinear interpolation
    this means we have to visit 2**N points. If N is known this is most
    clearly expressed as a set of N nested loops:

    J=-1
    for i1 in range(0,2):
        for i2 in range(0,2):
            ...
                for iN in range(0,2):
                    J+=1
                    do stuff for J'th item.

    Here instead, since we do not know N in advance, use a set of recursive
    loops to depth N, where at the final level we yield for each of the 2**N
    points the values of the points and the weights with which they contribute
    to the interpolated value.

    :param recursion_depth:
    :param indices_1d:
    :param weights_1d:
    :param narg: indices from outer recursive loops
    :return: generater function that yields the J"th thing to do stuff with.
    """
    number_of_coordinates = indices_1d.shape[0]
    number_of_points = indices_1d.shape[2]

    if recursion_depth > 0:
        # Loop over the n'th coordinate, with
        #    n = number_of_coordinates - recursion_depth
        for ii in range(0, 2):
            # Yield from next recursive loop, add the loop coordinate to the
            # arg of the next call
            arg = (*narg, ii)
            yield from _next_point(recursion_depth - 1, indices_1d, weights_1d, *arg)
    else:
        #
        # Here we construct the "fancy" indexes we will use to grab datavalues.
        indices_nd = []
        weights_nd = numpy.ones((number_of_points,), dtype="float64")
        for index in range(0, number_of_coordinates):
            # get the coordinate index for the current point.
            indices_nd.append(indices_1d[index, narg[index], :])

            # The N-dimensional weight is the multiplication of all weights
            # of the associated 1d problems
            weights_nd *= weights_1d[index, narg[index], :]

        yield indices_nd, weights_nd



import numpy
from numba import njit


def mem(
    directions_radians: numpy.ndarray,
    a1: numpy.ndarray,
    b1: numpy.ndarray,
    a2: numpy.ndarray,
    b2: numpy.ndarray,
    progress,
    **kwargs
) -> numpy.ndarray:

    number_of_frequencies = a1.shape[-1]
    number_of_points = a1.shape[0]

    directional_distribution = numpy.zeros(
        (number_of_points, number_of_frequencies, len(directions_radians))
    )

    for ipoint in range(0, number_of_points):
        progress.update(1)
        directional_distribution[ipoint, :, :] = _mem(
            directions_radians,
            a1[ipoint, :],
            b1[ipoint, :],
            a2[ipoint, :],
            b2[ipoint, :],
        )
    return directional_distribution


def _mem(
    directions_radians: numpy.ndarray,
    a1: numpy.ndarray,
    b1: numpy.ndarray,
    a2: numpy.ndarray,
    b2: numpy.ndarray,
) -> numpy.ndarray:
    """
    This function uses the maximum entropy method by Lygre and Krogstadt (1986,JPO)
    to estimate the directional shape of the spectrum. Enthropy is defined in the
    Boltzmann sense (log D)

    Lygre, A., & Krogstad, H. E. (1986). Maximum entropy estimation of the directional
    distribution in ocean wave spectra. Journal of Physical Oceanography, 16(12), 2052-2060.

    :param directions_radians: 1d array of wave directions in radians,
    length[number_of_directions]. (going to, anti-clockswise from east)

    :param a1: 1d array of cosine directional moment as function of frequency,
    length [number_of_frequencies]

    :param b1: 1d array of sine directional moment as function of frequency,
    length [number_of_frequencies]

    :param a2: 1d array of double angle cosine directional moment as function
    of frequency, length [number_of_frequencies]

    :param b2: 1d array of double angle sine directional moment as function of
    frequency, length [number_of_frequencies]

    :return: array with shape [number_of_frequencies,number_of_direction]
    representing the directional distribution of the waves at each frequency.

    Maximize the enthrophy of the solution with entrophy defined as:

           integrate log(D) over directions

    such that the resulting distribution D reproduces the observed moments.

    :return: Directional distribution as a numpy array

    Note that:
    d1 = a1; d2 =b1; d3 = a2 and d4=b2 in the defining equations 10.
    """

    number_of_directions = len(directions_radians)

    c1 = a1 + 1j * b1
    c2 = a2 + 1j * b2
    #
    # Eq. 13 L&K86
    #
    Phi1 = (c1 - c2 * numpy.conj(c1)) / (1 - c1 * numpy.conj(c1))
    Phi2 = c2 - Phi1 * c1
    #
    e1 = numpy.exp(-directions_radians * 1j)
    e2 = numpy.exp(-directions_radians * 2j)

    numerator = 1 - Phi1 * numpy.conj(c1) - Phi2 * numpy.conj(c2)
    denominator = (
        numpy.abs(1 - Phi1[:, None] * e1[None, :] - Phi2[:, None] * e2[None, :]) ** 2
    )

    D = numpy.real(numerator[:, None] / denominator) / numpy.pi / 2

    # Normalize to 1. in discrete sense
    integralApprox = numpy.sum(D, axis=-1) * numpy.pi * 2.0 / number_of_directions
    D = D / integralApprox[:, None]

    return numpy.squeeze(D)


@njit(cache=True)
def numba_mem(
    directions_radians: numpy.ndarray,
    a1: float,
    b1: float,
    a2: float,
    b2: float,
) -> numpy.ndarray:
    """
    This function uses the maximum entropy method by Lygre and Krogstadt (1986,JPO)
    to estimate the directional shape of the spectrum. Enthropy is defined in the
    Boltzmann sense (log D)

    Lygre, A., & Krogstad, H. E. (1986). Maximum entropy estimation of the directional
    distribution in ocean wave spectra. Journal of Physical Oceanography, 16(12), 2052-2060.

    :param directions_radians: 1d array of wave directions in radians,
    length[number_of_directions]. (going to, anti-clockswise from east)

    :param a1: 1d array of cosine directional moment as function of frequency,
    length [number_of_frequencies]

    :param b1: 1d array of sine directional moment as function of frequency,
    length [number_of_frequencies]

    :param a2: 1d array of double angle cosine directional moment as function
    of frequency, length [number_of_frequencies]

    :param b2: 1d array of double angle sine directional moment as function of
    frequency, length [number_of_frequencies]

    :return: array with shape [number_of_frequencies,number_of_direction]
    representing the directional distribution of the waves at each frequency.

    Maximize the enthrophy of the solution with entrophy defined as:

           integrate log(D) over directions

    such that the resulting distribution D reproduces the observed moments.

    :return: Directional distribution as a numpy array

    Note that:
    d1 = a1; d2 =b1; d3 = a2 and d4=b2 in the defining equations 10.
    """

    number_of_directions = len(directions_radians)

    c1 = a1 + 1j * b1
    c2 = a2 + 1j * b2
    #
    # Eq. 13 L&K86
    #
    Phi1 = (c1 - c2 * numpy.conj(c1)) / (1 - c1 * numpy.conj(c1))
    Phi2 = c2 - Phi1 * c1
    #
    e1 = numpy.exp(-directions_radians * 1j)
    e2 = numpy.exp(-directions_radians * 2j)

    numerator = 1 - Phi1 * numpy.conj(c1) - Phi2 * numpy.conj(c2)
    denominator = numpy.abs(1 - Phi1 * e1 - Phi2 * e2) ** 2

    D = numpy.real(numerator / denominator) / numpy.pi / 2

    # Normalize to 1. in discrete sense
    integralApprox = numpy.sum(D, axis=-1) * numpy.pi * 2.0 / number_of_directions
    D = D / integralApprox

    return D



import json
import os
import pandas
from typing import Literal, Callable
from datetime import datetime
from glob import glob
from typing import Iterator, Mapping
from numpy import timedelta64, zeros, float64, full
from numpy.typing import NDArray
from pandas import DataFrame, concat
from roguewave import to_datetime_utc
from roguewave.tools.time import to_datetime64
import xarray

CSV_TYPES = Literal[
    "FLT", "SPC", "GPS", "GMT", "LOC", "BARO", "BARO_RAW", "SST", "RAINDB"
]


# Main Functions
# ---------------------------------


def read_and_concatenate_spotter_csv(
    path,
    csv_type: CSV_TYPES,
    start_date: datetime = None,
    end_date: datetime = None,
    cache_as_netcdf=False,
) -> pandas.DataFrame:
    """
    Read data for a given data type from the given path.

    :param path: Path containing Spotter CSV files

    :param csv_type: One of the supported CSV file formats.

    :param start_date: If only a subset of the data is needed we can avoid
                       loading all data, this denotes the start date of the
                       desired interval. If given, only data after the
                       start_date is loaded (if available).
                       NOTE: this requires that LOC files are present.

    :param end_date: If only a subset of the data is needed we can avoid
                     loading all data, this denotes the end date of the
                     desired interval. If given, only data before the
                     end_date is loaded (if available).
                     NOTE: this requires that LOC files are present.
    """
    if cache_as_netcdf:
        if os.path.exists(_netcdf_filename(path, csv_type)):
            dataset = xarray.open_dataset(_netcdf_filename(path, csv_type))
            return dataset.to_dataframe()

    csv_format = get_csv_file_format(csv_type)
    source_files = list(
        _files_to_parse(path, csv_format["pattern"], start_date, end_date)
    )

    if len(source_files) == 0:
        raise FileNotFoundError("No files to parse")

    data_frames = [
        _process_file(source_file, csv_format) for source_file in source_files
    ]

    # Fragmentation occurs for spectral data- to avoid performance issues we recreate the dataframe after the concat
    # here.
    dataframe = pandas.concat(data_frames).copy()
    dataframe.reset_index(inplace=True)
    dataframe = _mark_continuous_groups(
        dataframe, csv_format["sampling_interval_seconds"]
    )

    # Restrict to reequested start and end date (if applicable).
    if start_date is not None:
        start_date = to_datetime64(start_date)
        dataframe = dataframe[dataframe["time"] >= start_date]

    if end_date is not None:
        end_date = to_datetime64(end_date)
        dataframe = dataframe[dataframe["time"] < end_date]

    if cache_as_netcdf:
        save_as_netcdf(dataframe, path, csv_type)

    return dataframe


# Private Functions
# ---------------------------------
def _files_to_parse(
    path: str, pattern: str, start_date: datetime = None, end_date: datetime = None
) -> Iterator[str]:
    """
    Create a generator that yields files that conform to a given filename pattern and that contain data within the
    requested interval. Returns an iterator.

    :param path: path containing Spotter data files
    :param pattern: file pattern to look for
    :param start_date: start date of the interval, if none no lower interval is imposed.
    :param end_date: end date of the interval, if none no upper interval is imposed.
    :return: generator that yields files qith data conforming to the request.
    """

    # Get the files conforming to a specific pattern- say ????_LOC.csv
    # For Spotter these are the location files, 0000_LOC.csv, 0001_LOC.csv, etc.
    files = glob(os.path.join(path, pattern))

    time_file_format = get_csv_file_format("TIME")
    if (start_date is not None) or (end_date is not None):
        # If we are only interested in a range of data we do not want to parse all the text files. Specifically the
        # location, gps and spectral files can get very large (hundreds of mb) and consequenlty slow to read. Spotter
        # saves this data onboard split across multiple files (e.g. 0000_FLT.csv, 0001_FLT.csv etc) so we could just
        # load only those files containing the time range we need. We do not in general know which files contain which
        # time range though. Here we there try to read the timebase from the smallest data files (location files)
        # and use that to determine which files we read.

        # Read the files containing the timebase.
        time_files = glob(os.path.join(path, time_file_format["pattern"]))
        csv_parsing_options = get_csv_file_format("TIME")

        if not (len(time_files) == len(files)):
            # We need an equal number of files - otherwise something is amiss
            raise ValueError(
                "Number of location files needs to be the same as the number of files being read to use"
                "time filtering"
            )

        # For each data file and time base file do
        for file, time_file in zip(sorted(files), sorted(time_files)):

            # Read the time base file to get the time range in the XXXX_VAR.csv file
            time = pandas.read_csv(
                time_file,
                index_col=False,
                delimiter=",",
                header=0,
                names=[x["column_name"] for x in csv_parsing_options],
                dtype={x["column_name"]: x["dtype"] for x in csv_parsing_options},
                on_bad_lines="skip",
                usecols=[x["column_name"] for x in csv_parsing_options if x["include"]],
            )
            file_in_range = False
            if time.shape[0] == 0:
                # If empty- continue
                continue

            # If a start date is given- check if any data in the file has a timestamp after the start date
            if start_date is not None:
                start_date_timestamp = to_datetime_utc(start_date).timestamp()
                file_in_range = not (
                    start_date_timestamp > time["time"].values[-1] + 60.0
                )

            # If an end date is given- check if any data in the file has a timestamp before the end date
            if end_date is not None:
                end_date_timestamp = to_datetime_utc(end_date).timestamp()
                file_in_range = file_in_range and not (
                    end_date_timestamp <= time["time"].values[0]
                )

            if file_in_range:
                yield file
            else:
                continue

    else:
        for file in sorted(files):
            yield file


def _process_file(file, csv_format, nrows=None):
    """
    Process a _single_ csv file of the form ????_TYPE.csv

    :param file: file name
    :param csv_format: the format of the csv file
    :param nrows: how many rows to read (default:None implies all rows)
    :return: dataframe containing date in the CSV file after quality control (filtering of nan rows etc.)
    """

    if not csv_format.get("ragged", False):
        df = pandas.read_csv(
            file,
            index_col=False,
            delimiter=",",
            skiprows=1,
            names=[x["name"] for x in csv_format["columns"]],
            on_bad_lines="skip",
            low_memory=False,
            nrows=nrows,
        )

    else:
        # Some spotter files are ragged (e.g. GMN) and may contain a variable number of columns. To handle
        # this we need to use the python parser.
        df = pandas.read_csv(
            file,
            index_col=False,
            delimiter=",",
            skiprows=1,
            names=[x["name"] for x in csv_format["columns"]],
            on_bad_lines="skip",
            engine="python",
            nrows=nrows,
        )

    # Assign the designated time column.
    df["time"] = df[csv_format["time_column"]]

    # If the time is in millis (milliseconds since system start) convert to unix epoch time in seconds (still float type
    # here)
    if csv_format["time_column"] == "millis":
        df["time"] = _milis_to_epoch(df["time"], file)

    # Convert the columns into the correct type. Coerce bad values to NaN values for numeric types.
    # (to note this is the reason we want to define numerics as floats, since integer do not contain a natural missing
    # value type)
    for column in csv_format["columns"]:
        name = column["name"]
        if column["dtype"] == "str":
            # Fill missing values in strings with blanks. Some Spotter files (FLT) contain a trailing string column
            # that does not always contain data and that gets filled with NA values. If we do not replace these with
            # blanks the subsequent dropping on NaN's in quality control would filter good data.
            df[name] = df[name].astype("string").fillna("")

        else:
            # TODO: what to do with malformed data in integer columns???
            df[name] = pandas.to_numeric(df[name], errors="coerce")

    # Drop NaN rows and remove any instances where we go back in time (only needed for raw GPS files).
    df = _quality_control(df, csv_format.get("dropna", True))

    # Convert time to proper pandas datetime format
    df["time"] = pandas.to_datetime(df["time"], unit="s")

    columns_to_drop = csv_format.get("drop", [])
    if columns_to_drop:
        df = df.drop(columns_to_drop, axis=1)

    return df


# Utility Functions
# ---------------------------------
def _netcdf_filename(path: str, csv_type: str) -> str:
    return os.path.join(path, f"{csv_type}.nc")


def save_as_netcdf(df: DataFrame, path, csv_type: CSV_TYPES) -> None:
    dataset = xarray.Dataset.from_dataframe(df)
    dataset.to_netcdf(_netcdf_filename(path, csv_type))


def _quality_control(dataframe: pandas.DataFrame, dropna) -> pandas.DataFrame:
    # Remove any rows with nan entries

    if dropna:
        dataframe = dataframe.dropna()

    # Here we check for negative intervals and only keep data with positive intervals. It needs to be recursively
    # applied since when removing an interval we may introduce a new negative interval. There are definitely more
    # efficient ways to implement this and if this ever is a bottleneck we should - until then we do it the ugly way :-)

    if "time" not in dataframe:
        return dataframe

    if dataframe.shape[0] < 1:
        return dataframe

    while True:
        positive_time = dataframe["time"].diff() > 0.0
        positive_time[0] = True
        if all(positive_time.values[1:]):
            break
        else:
            dataframe = dataframe.loc[positive_time]

    return dataframe


def _mark_continuous_groups(df: pandas.DataFrame, sampling_interval_seconds):
    """
    This function adds a column that has unique number for each continuous block of data (i.e. data without gaps).
    It does so by:
        - calculating the time difference
        - creating a boolean mask where the time difference is larger than a threshold
        - taking the cumulative sum over that mask- this sum mask will only increase if there is a gap between succesive
          entries - and can thus serve as a group marker.
    :param df: data frame contaning a time epoch key
    :param sampling_interval: sampling interval used
    :return:
    """

    if "time" in df:
        sampling_interval_seconds = timedelta64(
            int(sampling_interval_seconds * 1e9), "ns"
        )
        df["group id"] = (
            df["time"].diff() > sampling_interval_seconds + timedelta64(100, "ms")
        ).cumsum()
    else:
        df["group id"] = zeros(df.shape[0], dtype="int64")
    return df


def get_csv_file_format(csv_type: CSV_TYPES) -> Mapping:
    """
    Loads the CSV file format description from JSON files into a dict. All file formats
    are stored as JSON files in the ./file_formats directory.

    :param csv_type: which CSV format file to load.
    :return:
    """
    with open(
        os.path.join(os.path.dirname(__file__), f"file_formats/{csv_type}.json")
    ) as file:
        format = json.loads(file.read())
    return format


def _milis_to_epoch(millis: NDArray[float64], millis_file: str) -> NDArray[float64]:
    """
    Convert Spotter "millis" into unix epoch time stamp. Millis measure time elapsed in miliseconds since the start of
    the system. To translate these into a timestamp referenced to the Unix epoch we use the FLT files- since these
    contain both a Millis and Epoch time timestamp allowing us to determine the delta.

    :param millis: array of millis
    :param millis_file: filename of the array that contained the millis (used to get the proper FLT file)
    :return: array of corresponding epoch times
    """

    #
    if len(millis) == 0:
        # If len 0 there is nothing to convert
        return millis

    # get the path and filename of the milis file, and use it to deduce the corresponding
    # _FLT filename
    path, filename = os.path.split(millis_file)
    flt_file = os.path.join(path, filename.split("_")[0] + "_FLT.csv")

    # Get the data format for the flt file, and load 1 line of data from it
    csv_format = get_csv_file_format("FLT")
    data = _process_file(flt_file, csv_format, 1)

    # Use the millis and epoch time to deduce the systems epoch (aka startup) time.
    epoch = data[csv_format["time_column"]].values[0] - data["millis"].values[0] / 1000

    # Add the epoch to millis converted to seconds and return result.
    return epoch + millis / 1000


def apply_to_group(function: Callable[[DataFrame], DataFrame], dataframe: DataFrame):
    """
    Apply a function to each group seperately and recombine the result into a single dataframe.

    :param function: Function to appply
    :param dataframe: Dataframe to apply function to
    :return:
    """
    groups = dataframe.groupby("group id")

    dataframes = []
    for group_id, df in groups:
        _df = function(df)
        if "group id" not in _df:
            _df["group id"] = full(_df.shape[0], group_id)
        dataframes.append(_df)

    return concat(dataframes)



from .keygeneration import generate_uris

from roguewave.modeldata.timebase import (
    TimeSliceLead,
    timebase_forecast,
    TimeSliceForecast,
    TimeSliceEvaluation,
    TimeSliceAnalysis
)

from roguewave.modeldata.open_remote import open_remote_dataset
from roguewave.modeldata.extract import extract_from_remote_dataset
from roguewave.modeldata.keygeneration import generate_uris

from roguewave.modeldata.modelinformation import (
    model_timebase,
    model_valid_time,
    available_models,
    list_available_variables
)



from roguewave.wavephysics.balance import WindGeneration, Dissipation
from typing import Mapping, Tuple
from roguewave import FrequencyDirectionSpectrum
from roguewave.wavespectra.operations import numba_integrate_spectral_data
from roguewave.wavephysics.balance.generation import _u10_from_bulk_rate_point
from roguewave.wavephysics.balance.dissipation import _bulk_dissipation_direction_point
from xarray import DataArray, Dataset
from numba_progress import ProgressBar
from numba.typed import List as NumbaList
from numba import njit, prange
from numpy.typing import NDArray
from numpy import empty, isnan, zeros


class SourceTermBalance:
    def __init__(self, generation: WindGeneration, disspipation: Dissipation):
        self.generation = generation
        self.dissipation = disspipation

    def evaluate_imbalance(
        self,
        wind_speed: DataArray,
        wind_direction: DataArray,
        spectrum: FrequencyDirectionSpectrum,
        time_derivative_spectrum: FrequencyDirectionSpectrum = None,
    ) -> DataArray:

        if time_derivative_spectrum is None:
            time_derivative_spectrum = 0.0
        else:
            time_derivative_spectrum = time_derivative_spectrum.variance_density

        return (
            self.generation.rate(
                spectrum,
                wind_speed,
                wind_direction,
            )
            + self.dissipation.rate(spectrum)
            - time_derivative_spectrum
        )

    def evaluate_bulk_imbalance(
        self,
        wind_speed: DataArray,
        wind_direction: DataArray,
        spectrum: FrequencyDirectionSpectrum,
        time_derivative_spectrum: FrequencyDirectionSpectrum = None,
    ) -> DataArray:

        if time_derivative_spectrum is None:
            time_derivative_spectrum = 0.0
        else:
            time_derivative_spectrum = time_derivative_spectrum.m0()

        return (
            self.generation.bulk_rate(spectrum, wind_speed, wind_direction)
            + self.dissipation.bulk_rate(spectrum)
            - time_derivative_spectrum
        )

    def update_parameters(self, parameters: Mapping):
        for key in parameters:
            if key in self.generation._parameters:
                self.generation._parameters[key] = parameters[key]
            elif key in self.dissipation._parameters:
                self.dissipation._parameters[key] = parameters[key]

    def windspeed_and_direction_from_spectra(
        self,
        guess_u10: DataArray,
        spectrum: FrequencyDirectionSpectrum,
        jacobian=False,
        jacobian_parameters=None,
        time_derivative_spectrum: FrequencyDirectionSpectrum = None,
    ) -> Dataset:
        """

        :param bulk_rate:
        :param guess_u10:
        :param guess_direction:
        :param spectrum:
        :return:
        """
        disable = spectrum.number_of_spectra < 100
        if time_derivative_spectrum is None:
            time_derivative_spectrum = zeros(spectrum.shape())
        else:
            time_derivative_spectrum = time_derivative_spectrum.variance_density.values

        with ProgressBar(
            total=spectrum.number_of_spectra,
            disable=disable,
            desc=f"Estimating U10 from {self.generation.name} and {self.dissipation.name} "
            f"wind and dissipation source terms",
        ) as progress_bar:
            if not jacobian:
                speed, direction = _u10_from_spectra(
                    variance_density=spectrum.variance_density.values,
                    guess_u10=guess_u10.values,
                    depth=spectrum.depth.values,
                    wind_source_term_function=self.generation._wind_source_term_function,
                    dissipation_source_term_function=self.dissipation._dissipation_function,
                    parameters_generation=self.generation.parameters,
                    parameters_dissipation=self.dissipation.parameters,
                    spectral_grid=self.generation.spectral_grid(spectrum),
                    progress_bar=progress_bar,
                    time_derivative_spectrum=time_derivative_spectrum,
                )
            else:
                if jacobian_parameters is None:
                    raise ValueError(
                        "If gradients are requested a parameter list is required"
                    )

                speed, direction, grad = _u10_from_spectra_gradient(
                    variance_density=spectrum.variance_density.values,
                    guess_u10=guess_u10.values,
                    depth=spectrum.depth.values,
                    wind_source_term_function=self.generation._wind_source_term_function,
                    dissipation_source_term_function=self.dissipation._dissipation_function,
                    grad_parameters=NumbaList(jacobian_parameters),
                    parameters_generation=self.generation.parameters,
                    parameters_dissipation=self.dissipation.parameters,
                    spectral_grid=self.generation.spectral_grid(spectrum),
                    time_derivative_spectrum=time_derivative_spectrum,
                    progress_bar=progress_bar,
                )
                grad = DataArray(data=grad)

        u10 = DataArray(
            data=speed,
            dims=spectrum.dims_space_time,
            coords=spectrum.coords_space_time,
        )

        direction = DataArray(
            data=speed,
            dims=spectrum.dims_space_time,
            coords=spectrum.coords_space_time,
        )

        if jacobian:
            return Dataset(
                data_vars={"u10": u10, "direction": direction, "jacobian": grad}
            )
        else:
            return Dataset(data_vars={"u10": u10, "direction": direction})


# ----------------------------------------------------------------------------------------------------------------------
# Apply to all spatial points
# ----------------------------------------------------------------------------------------------------------------------


@njit(parallel=True, nogil=True)
def _u10_from_spectra(
    variance_density: NDArray,
    guess_u10: NDArray,
    depth: NDArray,
    wind_source_term_function,
    dissipation_source_term_function,
    parameters_generation,
    parameters_dissipation,
    spectral_grid,
    progress_bar: ProgressBar = None,
    time_derivative_spectrum=None,
) -> Tuple[NDArray, NDArray]:
    """

    :param variance_density:
    :param guess_u10:
    :param depth:
    :param wind_source_term_function:
    :param dissipation_source_term_function:
    :param parameters_generation:
    :param parameters_dissipation:
    :param spectral_grid:
    :param progress_bar:
    :return:
    """
    number_of_points = variance_density.shape[0]
    u10 = empty((number_of_points))
    direction = empty((number_of_points))

    for point_index in prange(number_of_points):
        if progress_bar is not None:
            progress_bar.update(1)

        u10[point_index], direction[point_index] = _u10_from_spectra_point(
            variance_density[point_index, :, :],
            guess_u10[point_index],
            depth[point_index],
            wind_source_term_function,
            dissipation_source_term_function,
            parameters_generation,
            parameters_dissipation,
            spectral_grid,
            time_derivative_spectrum[point_index, :, :],
        )
    return u10, direction


@njit(parallel=False)
def _u10_from_spectra_point(
    variance_density,
    guess_u10,
    depth,
    wind_source_term_function,
    dissipation_source_term_function,
    parameters_generation,
    parameters_dissipation,
    spectral_grid,
    time_derivative_spectrum,
) -> Tuple[float, float]:
    """

    :param bulk_rate:
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param guess_u10:
    :param depth:
    :param parameters:
    :param spectral_grid:
    :param progress_bar:
    :return:
    """

    direction, bulk_rate = _bulk_dissipation_direction_point(
        variance_density,
        depth,
        dissipation_source_term_function,
        spectral_grid,
        parameters_dissipation,
    )

    # Note dissipation is negatve- but our target bulk wind generation is positive
    u10 = _u10_from_bulk_rate_point(
        -bulk_rate,
        variance_density,
        guess_u10,
        direction,
        depth,
        spectral_grid,
        parameters_generation,
        wind_source_term_function,
        time_derivative_spectrum,
    )
    return u10, direction


@njit(parallel=False, cache=True)
def _u10_parameter_gradient(
    variance_density,
    guess_u10,
    depth,
    wind_source_term_function,
    dissipation_source_term_function,
    grad_parameters,
    parameters_generation,
    parameters_dissipation,
    spectral_grid,
    time_derivative_spectrum,
) -> (float, float, NDArray):
    """
    Function to numerically calculate gradients for the requested coeficients
    :param bulk_rate:
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param guess_u10:
    :param depth:
    :param parameters:
    :param spectral_grid:
    :param progress_bar:
    :return:
    """

    # Calculate the zero point
    direction, bulk_rate = _bulk_dissipation_direction_point(
        variance_density,
        depth,
        dissipation_source_term_function,
        spectral_grid,
        parameters_dissipation,
    )

    # Note dissipation is negatve- but our target bulk wind generation is positive
    u10 = _u10_from_bulk_rate_point(
        -bulk_rate,
        variance_density,
        guess_u10,
        direction,
        depth,
        spectral_grid,
        parameters_generation,
        wind_source_term_function,
        time_derivative_spectrum,
    )
    grad = empty(len(grad_parameters))
    if isnan(u10):
        grad[:] = 0
        return u10, direction, grad

    for index, param in enumerate(grad_parameters):
        perturbed_parameters_dissipation = parameters_dissipation.copy()
        perturbed_parameters_generation = parameters_generation.copy()

        if param in parameters_dissipation:
            step = 0.05 * abs(perturbed_parameters_dissipation[param])
            perturbed_parameters_dissipation[param] += step

            dissipation = dissipation_source_term_function(
                variance_density=variance_density,
                depth=depth,
                spectral_grid=spectral_grid,
                parameters=perturbed_parameters_dissipation,
            )

            new_bulk_rate = numba_integrate_spectral_data(dissipation, spectral_grid)

        else:
            step = 0.05 * abs(perturbed_parameters_generation[param])
            perturbed_parameters_generation[param] += step
            new_bulk_rate = bulk_rate

        new_u10 = _u10_from_bulk_rate_point(
            -new_bulk_rate,
            variance_density,
            u10,
            direction,
            depth,
            spectral_grid,
            perturbed_parameters_generation,
            wind_source_term_function,
            time_derivative_spectrum,
        )
        if isnan(new_u10):
            grad[index] = 0
        else:
            grad[index] = (new_u10 - u10) / step

    return u10, direction, grad


@njit(parallel=True)
def _u10_from_spectra_gradient(
    variance_density,
    guess_u10,
    depth,
    wind_source_term_function,
    dissipation_source_term_function,
    grad_parameters,
    parameters_generation,
    parameters_dissipation,
    spectral_grid,
    time_derivative_spectrum,
    progress_bar: ProgressBar = None,
) -> Tuple[NDArray, NDArray, NDArray]:
    """

    :param variance_density:
    :param guess_u10:
    :param depth:
    :param wind_source_term_function:
    :param dissipation_source_term_function:
    :param parameters_generation:
    :param parameters_dissipation:
    :param spectral_grid:
    :param progress_bar:
    :return:
    """
    number_of_points = variance_density.shape[0]
    u10 = empty((number_of_points))
    direction = empty((number_of_points))
    grad = empty((number_of_points, len(grad_parameters)))
    for point_index in prange(number_of_points):
        if progress_bar is not None:
            progress_bar.update(1)

        (
            u10[point_index],
            direction[point_index],
            grad[point_index, :],
        ) = _u10_parameter_gradient(
            variance_density[point_index, :, :],
            guess_u10[point_index],
            depth[point_index],
            wind_source_term_function,
            dissipation_source_term_function,
            grad_parameters,
            parameters_generation,
            parameters_dissipation,
            spectral_grid,
            time_derivative_spectrum,
        )
    return u10, direction, grad



# Author: Pieter Bart Smit

"""
A set of spectral analysis routines specifically designed to work well with data
from Sofar Spotter ocean wave buoys.

Classes:

- `SpectralAnalysisConfig`, configuration object

Functions:

- `segment_timeseries`, split timeseries into non-overlapping segments
- `calculate_moments`, calculate directional moments from (co)variances
- `spectral_analysis`, calculate wave spectrum for specific segment
- `generate_spectra`, calculate wave spectra for all segments
- `window_power_correction_factor`, correct for power loss due to windowing

"""

from numba import njit, objmode, types
from numba.typed import Dict
import numpy
from numpy.fft import fft
from numpy.typing import NDArray
from roguewave import FrequencySpectrum
from scipy.signal.windows import get_window
from typing import List, Tuple, Mapping
from xarray import Dataset


def estimate_frequency_spectrum(
    epoch_time: NDArray,
    x: NDArray,
    y: NDArray,
    z: NDArray,
    window: NDArray = None,
    segment_length_seconds=1800,
    sampling_frequency=2.5,
    options=None,
    spectral_window=None,
    response_functions=None,
    **kwargs
) -> FrequencySpectrum:
    """

    :param epoch_time:
    :param x:
    :param y:
    :param z:
    :param window:
    :param segment_length_seconds:
    :param sampling_frequency:
    :param options:
    :param spectral_window:
    :param kwargs:
    :return:
    """

    if window is None:
        window = get_window("hann", 256)

    options = _to_numba_dict(options)

    spectral_time, frequencies, co_spectra = estimate_co_spectra(
        epoch_time,
        (x, y, z),
        window,
        segment_length_seconds,
        sampling_frequency,
        options,
        spectral_window,
    )

    if response_functions is not None:
        for m in range(0, 3):
            for n in range(m, 3):
                response = response_functions[m](frequencies) * numpy.conjugate(
                    response_functions[n](frequencies)
                )
                co_spectra[:, m, n, :] = co_spectra[:, m, n, :] * response

                if m != n:
                    co_spectra[:, n, m, :] = co_spectra[:, m, n, :]

    szz, a1, b1, a2, b2 = _extract_moments(co_spectra, index_x=0, index_y=1, index_z=2)

    dataset = Dataset(
        data_vars={
            "variance_density": (["time", "frequency"], szz),
            "a1": (["time", "frequency"], a1),
            "b1": (["time", "frequency"], b1),
            "a2": (["time", "frequency"], a2),
            "b2": (["time", "frequency"], b2),
            "depth": (["time"], numpy.full(szz.shape[0], numpy.inf)),
            "latitude": (["time"], numpy.full(szz.shape[0], numpy.nan)),
            "longitude": (["time"], numpy.full(szz.shape[0], numpy.nan)),
        },
        coords={"time": spectral_time.astype("<M8[s]"), "frequency": frequencies},
    )
    return FrequencySpectrum(dataset)


@njit(cache=True)
def estimate_co_spectra(
    epoch_time: numpy.ndarray,
    signals,
    window: numpy.ndarray,
    segment_length_seconds,
    sampling_frequency,
    options: Dict = None,
    spectral_window=None,
) -> Tuple[NDArray, NDArray, NDArray]:
    segments = _segment_timeseries(epoch_time, segment_length_seconds)

    number_of_spectra = len(segments)
    number_of_frequencies = len(window) // 2

    nsig = len(signals)
    co_spectra = numpy.empty(
        (number_of_spectra, nsig, nsig, number_of_frequencies), dtype="complex_"
    )
    spectral_time = numpy.empty(number_of_spectra)

    for index, segment in enumerate(segments):
        istart = segment[0]
        iend = segment[1]
        spectral_time[index] = (epoch_time[istart] + epoch_time[iend]) / 2

        segment_signal = [sig[istart:iend] for sig in signals]
        co_spectra[index, :, :, :] = calculate_co_spectra(
            epoch_time[istart:iend],
            segment_signal,
            window,
            sampling_frequency,
            options,
            spectral_window,
        )

    df = sampling_frequency / len(window)
    frequencies = (
        numpy.linspace(numpy.int64(0), len(window) // 2 - 1, len(window) // 2) * df
    )

    return spectral_time, frequencies, co_spectra


# --------------------------
# Main spectral estimator
# --------------------------


@njit(cache=True)
def calculate_co_spectra(
    time_seconds: NDArray,
    signals: Tuple[NDArray],
    window: NDArray,
    sampling_frequency: float,
    options: Dict = None,
    spectral_window: NDArray = None,
) -> NDArray:
    """
    Calculate co-spectral density matrix using Welch's method.

    :param time_seconds:
    :param signals:
    :param window:
    :param sampling_frequency:
    :param options:
    :param spectral_window:
    :return:
    """

    if options is None:
        options = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

    window_overlap_fraction = options.get("window_overlap_fraction", 0.5)
    timebase_jitter_fraction = options.get("timebase_jitter_fraction", 0.01)

    # STEP 0: Preliminaries
    # ---------------------------------------------------------------------------

    # Signal and window lengths
    overlap = int(len(window) * window_overlap_fraction)
    nyquist_index = len(window) // 2
    nsig = len(signals)
    nt = len(signals[0])

    # Initialize output and work arrays
    output = numpy.zeros((nsig, nsig, nyquist_index), dtype="complex_")
    ffts = numpy.zeros((nsig, nyquist_index), dtype="complex_")
    zero_mean_signals = numpy.zeros((nsig, nt))

    # Correct the signals for mean contribution (detrend)
    for index, signal in enumerate(signals):
        zero_mean_signals[index, :] = signal - numpy.nanmean(signal)

    # Scale the time so that it runs as [0,1,...]. Note, while we may have some jitter (not exactly integeres) here
    # it is assumed there are no gaps in the signal.
    time_base = (time_seconds - time_seconds[0]) * sampling_frequency
    time_delta = numpy.diff(time_base)

    # initialize counters
    number_of_realizations = 0  # Number of valid realizations that fit in the signal
    istart = 0  # Start index of the window

    # STEP 1: Calculate windowed raw (unscaled) co-spectra
    # ---------------------------------------------------------------------------

    while True:

        # Find the end of the window
        iend = istart + len(window)

        # If the end index is larger than the length of the signal we are done
        if iend > nt:
            break

        # check if there are any nan's in the signal signalling that this part of the signal is to be ignored in the
        # spectral analysis (e.g. due to quality control).
        nan_in_signal = False
        for index in range(nsig):
            if numpy.any(numpy.isnan(zero_mean_signals[index, istart:iend])):
                nan_in_signal = True
                break

        # If there are any time delta's larger than the sample time step we reject the realization
        if (
            numpy.any(time_delta[istart:iend] > 1 + timebase_jitter_fraction)
            or nan_in_signal
        ):
            istart = istart + len(window) - overlap
            continue

        # We have a "valid" realization
        number_of_realizations += 1

        # Calculate the fft of each of the signals
        for index in range(nsig):
            with objmode():
                # FFT not supported in Numba- yet
                ffts[index, :] = fft(zero_mean_signals[index, istart:iend] * window)[
                    0:nyquist_index
                ]

        # Calculate the co(spectral) contributions, making use of the Hermetian property of the co-spectral matrix,
        # i.e.   cospectra[mm,nn] == conjugate( cospectra[nn,mm] )
        for mm in range(0, nsig):
            for nn in range(mm, nsig):
                output[mm, nn, :] += 2 * ffts[mm, :] * numpy.conjugate(ffts[nn, :])

                if mm != nn:
                    output[nn, mm, :] = numpy.conjugate(output[mm, nn, :])

        # Advance the window to the next position
        istart = istart + len(window) - overlap

    # If no viable windows were found - return nan's
    if number_of_realizations == 0:
        output[:, :, :] = numpy.nan
        return output

    # STEP 2: Scale the co-spectra to an energy density output
    # ---------------------------------------------------------------------------

    # Multiply with scaling coeficient to account for:
    #  - power loss due to windowing -> * _window_power_correction_factor(window) ** 2
    #  - taking the average over all realizations -> / number_of_realizations
    #  - conversion to power density ->  / frequency_step
    #  - Fourier scaling constant of the FFT ->  / len(window) ** 2
    frequency_step = sampling_frequency / len(window)
    output *= (
        _window_power_correction_factor(window) ** 2
        / number_of_realizations
        / frequency_step
        / len(window) ** 2
    )

    # STEP 3: Convolutional smoothing in the frequency domain
    # ---------------------------------------------------------------------------

    # Apply spectral smoothing if requested.
    if spectral_window is not None:
        # Convolutional padding length
        pad_len = len(spectral_window) // 2

        # Do the smoothing for each of the signals, making use of the Hermetian property for efficiency
        for mm in range(0, nsig):
            for nn in range(mm, nsig):

                total_energy_pre_smoothing = numpy.sum(output[mm, nn, :])
                if total_energy_pre_smoothing == 0.0:
                    # In case of no energy (mostly for artificial or filtered input signals) we leave the signl
                    # unaltered (our renormalization would otherwise introduce a 0/0 -> nan
                    continue

                # Window avereaged energy through convolution.
                smoothed = numpy.convolve(output[mm, nn, :], spectral_window)

                # Renormalization to ensure we have the same pre/post total complex (co)variance
                total_energy_post_smoothing = numpy.sum(smoothed[pad_len:-pad_len])
                renormalization_factor = (
                    total_energy_pre_smoothing / total_energy_post_smoothing
                )
                output[mm, nn, :] = renormalization_factor * smoothed[pad_len:-pad_len]

                # Update the opposite diagonal with the conjugate of the signal
                if mm != nn:
                    output[nn, mm] = numpy.conjugate(output[mm, nn])

    return output


# ------------------
# Helper Functions
# ------------------


def _to_numba_dict(mapping: Mapping) -> Dict:
    """
    Convert a regular dictionary to a numba dictionary we can pass to jitted functions. Note that numba only supports
    Dicts that have the same type for the keys and the values for all entries.

    Purpose: mostly to make it easy to pass a normal python dict and convert it so something we can pass to
             numba.

    :param mapping: mapping to turn into a dict.
    :return:
    """

    numba_mapping = Dict.empty(key_type=types.unicode_type, value_type=types.float64)

    if mapping is not None:
        for key, item in mapping.items():
            numba_mapping[key] = item

    return numba_mapping


def _extract_moments(
    co_spectra: NDArray, index_x: int, index_y: int, index_z: int
) -> Tuple[NDArray, NDArray, NDArray, NDArray, NDArray]:
    """
    Extract the wave directional moments from a given co-spectral Matrix. Returns a tuple of numpy arrays as
    (szz,a1,b1,a2,b2)

    :param co_spectra: co-spectral matrix that contains (co)spectra of displacements.
    :param index_x: index of x in co-spectral array
    :param index_y: index of y in co-spectral array
    :param index_z: index of z in co-spectral array
    :return: Directional moments (szz,a1,b1,a2,b2)
    """

    # Vertical, and horizontal power spectra.
    szz = numpy.real(co_spectra[:, index_z, index_z, :])
    syy = numpy.real(co_spectra[:, index_y, index_y, :])
    sxx = numpy.real(co_spectra[:, index_x, index_x, :])

    # Co and quad spectra.
    cxy = numpy.real(co_spectra[:, index_x, index_y, :])
    qzx = numpy.imag(co_spectra[:, index_z, index_x, :])
    qzy = numpy.imag(co_spectra[:, index_z, index_y, :])

    # Calculat the directional moments from displacement (co) spectra
    a1 = qzx / numpy.sqrt(szz * (sxx + syy))
    b1 = qzy / numpy.sqrt(szz * (sxx + syy))
    a2 = (sxx - syy) / (sxx + syy)
    b2 = 2 * cxy / (sxx + syy)

    return szz, a1, b1, a2, b2


@njit(cache=True)
def _window_power_correction_factor(window: numpy.ndarray) -> float:
    """
    Calculate the power correction factor that corrects for the loss in power
    due to the aplication of the windowing function.

    Basically for a DC signal we want the windowed mean power to be equal to the non-windowed power

    mean (   (1 * ScaledWindow) **2 )  = mean( 1 )

    whereas the raw window gives

    Power = mean (   (Window) **2 )

    Hence- ScaledWindow = Window * sqrt( 1 / mean(window**2) )

    :param window: Window as numpy ndarray
    :return:
    """
    return 1 / numpy.sqrt(numpy.mean(window**2))


@njit(cache=True)
def _segment_timeseries(epoch_time, segment_length_seconds) -> List[tuple]:
    """
    This function segments the time series into segments of the given length.
    It returns a list of tuples containing the indices of the start and end
    point of each segment. Segmens are open ended [t_start,t_end>.

    If the segment length does not fit a whole number of times into the vector
    the partial remaining segment is discarded.

    :param epoch_time: time vector (seconds, unix epoch)
    :param segment_length_seconds: lenght of segment in segments
    :return:
    """

    # Initialize the start time and the expected next segment start
    istart = 0
    t_start = epoch_time[0]
    t_next = t_start + segment_length_seconds

    segments = []
    # loop over all times
    for index, time in enumerate(epoch_time):
        # if the current time is larger or equal to the next segment start,
        # create a new segment
        if time > t_next:
            # Add segment to list
            segments.append((istart, index))

            # setup the new start time/index and next expected segment start
            istart = index
            t_start = t_next
            t_next = t_start + segment_length_seconds

    return segments



import numpy
from dataclasses import dataclass
from typing import Sequence, Mapping, List, Union
from pandas import DataFrame
from roguewave.tools.time import to_datetime64, to_datetime_utc
from datetime import datetime
from roguewave.interpolate.general import interpolate_periodic
from numbers import Number


class _Geometry:
    pass


@dataclass()
class SpatialPoint(_Geometry):
    latitude: float
    longitude: float
    id: str

    @property
    def is_valid(self) -> bool:
        return numpy.isfinite(self.latitude) and numpy.isfinite(self.longitude)


@dataclass()
class SpaceTimePoint(SpatialPoint):
    time: datetime

    @classmethod
    def from_spatial_point(cls, point: SpatialPoint, time: datetime):
        return cls(point.latitude, point.longitude, point.id, time)


@dataclass()
class Cluster(_Geometry):
    """
    A cluster is a set of points, each identified by unique id.
    """

    points: Mapping[str, SpatialPoint]

    @classmethod
    def from_lat_lon_arrays(cls, lats, lons):
        points = {}
        for index in range(0, len(lats)):
            _id = str(index)
            points[_id] = SpatialPoint(lats[index], lons[index], _id)

        return cls(points)

    @property
    def latitude(self) -> numpy.ndarray:
        return numpy.array([x.latitude for x in self.points.values()])

    @property
    def longitude(self) -> numpy.ndarray:
        return numpy.array([x.longitude for x in self.points.values()])

    @property
    def ids(self) -> Sequence[str]:
        return [x for x in self.points.keys()]


@dataclass()
class ClusterStack(_Geometry):
    """
    A cluster timestack is a stack of clusters in time, e.g. a cluster of
    spotters as it evolves in time.
    """

    time: numpy.ndarray
    clusters: Sequence[Cluster]

    def as_track_set(self) -> "TrackSet":
        return TrackSet.from_clusters(self)

    @classmethod
    def from_track_set(cls, track_set: "TrackSet", time):
        return track_set.as_cluster_time_stack(time)

    def __len__(self):
        return len(self.time)


class Track(_Geometry):
    """
    A track is  the drift track of a single buoy in time
    """

    def __init__(self, points: List[SpaceTimePoint], id):
        points = list(filter(lambda x: x.is_valid, points))
        self.points = points
        self.id = id

    @property
    def time(self) -> numpy.ndarray:
        return to_datetime64([x.time for x in self.points])

    @property
    def latitude(self) -> numpy.ndarray:
        return numpy.array([x.latitude for x in self.points])

    @property
    def longitude(self) -> numpy.ndarray:
        return numpy.array([x.longitude for x in self.points])

    def interpolate(self, target_time) -> "Track":
        target_time = to_datetime64(target_time)
        time = self.time
        latitude = interpolate_periodic(
            time.astype("float64"),
            self.latitude,
            target_time.astype("float64"),
            left=self.latitude[0],
            right=self.latitude[-1],
        )
        longitude = interpolate_periodic(
            time.astype("float64"),
            self.longitude,
            target_time.astype("float64"),
            left=self.longitude[0],
            right=self.longitude[-1],
            fp_period=360,
        )
        return self.from_arrays(latitude, longitude, target_time, self.id)

    @classmethod
    def from_arrays(cls, latitude, longitude, time, id) -> "Track":
        points = []
        for lat, lon, t in zip(latitude, longitude, time):
            points.append(SpaceTimePoint(lat, lon, id, to_datetime_utc(t)))
        return cls(points, id)

    @classmethod
    def from_spotter(cls, spotter_id, spotter):
        points = []
        if isinstance(spotter, DataFrame):
            for latitude, longitude, time in zip(
                spotter["latitude"].values,
                spotter["longitude"].values,
                spotter["time"].values,
            ):
                points.append(
                    SpaceTimePoint(
                        time=time, latitude=latitude, longitude=longitude, id=spotter_id
                    )
                )
        elif hasattr(spotter, "dataset"):
            for latitude, longitude, time in zip(
                spotter.dataset["latitude"].values,
                spotter.dataset["longitude"].values,
                spotter.dataset["time"].values,
            ):
                points.append(
                    SpaceTimePoint(
                        time=time, latitude=latitude, longitude=longitude, id=spotter_id
                    )
                )
        else:
            for x in spotter:
                points.append(
                    SpaceTimePoint(
                        time=x.time,
                        latitude=x.latitude,
                        longitude=x.longitude,
                        id=spotter_id,
                    )
                )
        return cls(points, spotter_id)

    def __len__(self):
        return len(self.points)


@dataclass()
class TrackSet(_Geometry):
    """
    A collection of tracks is a set of tracks for multiple buoys.
    """

    tracks: Mapping[str, Track]

    @classmethod
    def from_spotters(cls, spotters: Mapping):
        tracks = {}
        for spotter_id, spotter in spotters.items():
            tracks[spotter_id] = Track.from_spotter(spotter_id, spotter)
        return cls(tracks)

    @classmethod
    def from_clusters(cls, cluster_time_stack: "ClusterStack"):
        tracks = {}
        for cluster, time in zip(cluster_time_stack.clusters, cluster_time_stack.time):

            for _id, point in cluster.points:
                space_time_point = SpaceTimePoint(
                    longitude=point.latitude,
                    latitude=point.longitude,
                    time=time,
                    id=_id,
                )
                if _id not in tracks:
                    tracks[_id] = Track([space_time_point], _id)
                else:
                    tracks[_id].points.append(space_time_point)
        return cls(tracks)

    def as_cluster_time_stack(self, time):
        tracks = self.interpolate(time)

        clusters = [{} for x in range(len(time))]
        for _id, track in tracks.tracks.items():
            for ii, point in enumerate(track.points):
                if point.is_valid:
                    index = numpy.searchsorted(
                        time, [to_datetime64(point.time)], side="left"
                    )[0]

                    clusters[index][_id] = SpatialPoint(
                        point.latitude, point.longitude, _id
                    )

        clusters = [Cluster(x) for x in clusters]
        return ClusterStack(time, clusters)

    def interpolate(self, time) -> "TrackSet":
        time = to_datetime64(time)
        tracks = {}
        for id, track in self.tracks.items():
            tracks[id] = track.interpolate(time)
        return TrackSet(tracks)

    @classmethod
    def from_cluster(cls, cluster: Cluster, time: numpy.ndarray) -> "TrackSet":
        tracks = {}
        for _id, point in cluster.points.items():
            track_points = [SpaceTimePoint.from_spatial_point(point, t) for t in time]
            tracks[point.id] = Track(points=track_points, id=point.id)
        return TrackSet(tracks)


Geometry = Union[
    _Geometry, Sequence[numpy.ndarray], Sequence[Sequence], Sequence[Number], Mapping
]


def convert_to_cluster_stack(geometry: Geometry, time: numpy.ndarray) -> ClusterStack:

    if isinstance(geometry, TrackSet):
        return geometry.as_cluster_time_stack(time)

    elif isinstance(geometry, Cluster):
        return ClusterStack(time, [geometry for _ in time])

    elif isinstance(geometry, Track):
        tracks = TrackSet({"track": geometry})
        return ClusterStack.from_track_set(tracks, time)

    elif isinstance(geometry, ClusterStack):
        return geometry

    elif isinstance(geometry, Sequence):
        tracks = convert_to_track_set(geometry, time)
        return ClusterStack.from_track_set(tracks, time)

    elif isinstance(geometry, Mapping):
        tracks = convert_to_track_set(geometry, time)
        return ClusterStack.from_track_set(tracks, time)


def convert_to_track_set(geometry: Geometry, time: numpy.ndarray) -> TrackSet:

    if isinstance(geometry, TrackSet):
        return geometry.interpolate(time)

    elif isinstance(geometry, Cluster):
        return TrackSet.from_clusters(ClusterStack(time, [geometry for _ in time]))

    elif isinstance(geometry, Track):
        return TrackSet({"track": geometry.interpolate(time)})

    elif isinstance(geometry, ClusterStack):
        return geometry.as_track_set()

    elif isinstance(geometry, Sequence):
        if not (isinstance(geometry[0], Sequence) or isinstance(geometry[0], Sequence)):
            _geometry = [[geometry[0]], [geometry[1]]]
        else:
            _geometry = geometry

        cluster = Cluster.from_lat_lon_arrays(_geometry[0], _geometry[1])
        return TrackSet.from_cluster(cluster, time)

    elif isinstance(geometry, Mapping):
        return TrackSet.from_spotters(geometry).interpolate(time)



from pandas import DataFrame
from roguewave.interpolate.general import interpolate_periodic
from roguewave.tools.time import to_datetime64
import numpy


def interpolate_dataframe_time(dataframe: DataFrame, new_time: numpy.ndarray):

    output = DataFrame()
    output["time"] = new_time
    columns = list(dataframe.columns)
    old_time = to_datetime64(dataframe["time"].values)
    new_time = to_datetime64(new_time)

    for name in columns:
        name: str
        period = None
        if "direction" in name.lower():
            fp_discont = 360
            fp_period = 360
        else:
            fp_discont = None
            fp_period = None

        if name == "time":
            continue

        # Interpolation does not work on anything other than numeric types. If we have an object - we just ignore it
        # (do not include column in output). Fixes a crash due to the new "processing_source" adding a string to
        # Spotter Api data that descibes where the data was processed.
        if dataframe[name].dtype == numpy.dtype(object):
            print(name)
            continue

        output[name] = interpolate_periodic(
            old_time.astype("float64"),
            dataframe[name].values,
            new_time.astype("float64"),
            x_period=period,
            fp_period=fp_period,
            fp_discont=fp_discont,
        )
    return output



from numba import njit
from numpy import (
    empty_like,
    arctan,
    tan,
    sqrt,
    cos,
    sin,
    pi,
    exp,
    zeros_like,
    abs,
    array,
    zeros,
)
from numpy.typing import NDArray

DEFAULT_ORDER = 4
DEFAULT_N = 1


@njit(cache=True)
def integrate(
    time: NDArray, signal: NDArray, order=DEFAULT_ORDER, n=DEFAULT_N, start_value=0.0
) -> NDArray:
    """
    Cumulatively integrate the given discretely sampled signal in time using a Newton-Coases like formulation of
    requested order and layout. Note that higher order methods are only used in regions where the timestep is constant
    across the integration stencil- otherwise we fall back to the trapezoidal rule which can handle variable timesteps.
    A small amount of jitter (<1%) in timesteps is permitted though (and effectively ignored).

    NOTE: by default we start at 0.0 - which in general means that for a zero-mean process we will pick up a random
          offset that will need to be corracted afterwards. (out is not zero-mean).

    :param time: ndarray of length nt containing the elapsed time in seconds.
    :param signal: ndarray of length nt containing the signal to be integrated
    :param order: Order of the returned Newton-Coates integration approximation.
    :param n: number of future points in the integration stencil.
    :param start_value: Starting value of the integrated signal.
    :return: NDARRAY of length nt that contains the integrated signal that starts at the requested start_value.
    """

    primary_stencil = integration_stencil(order, n)
    primary_stencil_width = len(primary_stencil)

    integrated_signal = empty_like(signal)
    integrated_signal[0] = start_value
    integrated_signal[:] = 0.0

    number_of_constant_time_steps = 0
    prev_dt = time[1] - time[0]
    restart = True

    nt = len(signal)
    for ii in range(1, len(signal)):
        curr_dt = time[ii] - time[ii - 1]

        if ii + n - 1 < nt:
            future_dt = time[ii + n - 1] - time[ii + n - 2]
        else:
            future_dt = curr_dt
            restart = True

        if abs(future_dt - prev_dt) > 0.01 * curr_dt:
            # Jitter in the timestep, fall back to a lower order method that can handle this.
            restart = True
            number_of_constant_time_steps = 0

        if restart:
            number_of_constant_time_steps += 1
            stencil_width = 2
            stencil = array([0.5, 0.5])
            number_of_implicit_points = 1

            if number_of_constant_time_steps == primary_stencil_width:
                # We now have a series of enough constant timesteps to go to the higher order method.
                restart = False

        else:
            stencil_width = primary_stencil_width
            stencil = primary_stencil
            number_of_implicit_points = n

        delta = 0.0
        jstart = -(stencil_width - number_of_implicit_points)
        jend = number_of_implicit_points
        for jj in range(jstart, jend):
            delta += stencil[jj - jstart] * signal[ii + jj]

        integrated_signal[ii] = integrated_signal[ii - 1] + delta * curr_dt
        prev_dt = curr_dt

    return integrated_signal


@njit(cache=True)
def cumulative_distance(latitudes, longitudes):
    semi_major_axis = 6378137
    semi_minor_axis = 6356752.314245
    # eccentricity - squared
    eccentricity_squared = (
        semi_major_axis**2 - semi_minor_axis**2
    ) / semi_major_axis**2

    x = empty_like(latitudes)
    y = empty_like(longitudes)
    x[0] = 0
    y[0] = 0

    for ii in range(1, len(latitudes)):
        delta_longitude = (
            ((longitudes[ii] - longitudes[ii - 1] + 180) % 360 - 180) * pi / 180
        )
        delta_latitude = (latitudes[ii] - latitudes[ii - 1]) * pi / 180

        mean_latitude = (latitudes[ii] + latitudes[ii - 1]) / 2 * pi / 180

        # reduced latitude
        reduced_latitude = arctan(sqrt(1 - eccentricity_squared) * tan(mean_latitude))

        # length of a small meridian arc
        arc_length = (
            semi_major_axis
            * (1 - eccentricity_squared)
            * (1 - eccentricity_squared * sin(mean_latitude) ** 2) ** (-3 / 2)
        )

        x[ii] = x[ii - 1] + delta_longitude * semi_major_axis * cos(reduced_latitude)
        y[ii] = y[ii - 1] + arc_length * delta_latitude
    return x, y


@njit(cache=True)
def complex_response(
    normalized_frequency: NDArray, order: int, number_of_implicit_points: int = 1
):
    """
    The frequency complex response factor of the numerical integration scheme with given order and number of
    implicit points.

    :param normalized_frequency: Frequency normalized with the sampling frequency to calculate response factor at
    :param order: Order of the returned Newton-Coates integration approximation.
    :param number_of_implicit_points: number of future points in the integration stencil.
    :return: complex NDArray of same length as the input frequency containing the response factor at the given
             frequencies
    """
    number_of_explicit_points = order - number_of_implicit_points
    stencil = integration_stencil(order, number_of_implicit_points)

    normalized_omega = 2j * pi * normalized_frequency

    response_factor = zeros_like(normalized_frequency, dtype="complex_")
    for ii in range(-number_of_explicit_points, number_of_implicit_points):
        response_factor += stencil[ii + number_of_explicit_points] * exp(
            normalized_omega * ii
        )

    for index, omega in enumerate(normalized_omega):
        if omega == 0.0 + 0.0j:
            response_factor[index] = 1.0 + 0.0j
        else:
            response_factor[index] = response_factor[index] * (
                omega / (1 - exp(-omega))
            )

    return response_factor


@njit(cache=True)
def lagrange_base_polynomial_coef(order, base_polynomial_index):
    """
    We consider the interpolation of Y[0] Y[1] ... Y[order] spaced 1 apart at 0, 1,... point_index, ... order in terms
    of the Lagrange polynomial:

    Y[x]  =   L_0[x] Y[0] + L_1[x] Y[1] + .... L_order[x] Y[order].

    Here each of the lagrange polynomial coefficients L_n is expressed as a polynomial in x

    L_n = a_0 x**(order-1) + a_1 x**(order-2) + ... a_order

    where the coeficients may be found from the standard definition of the base polynomial (e.g. for L_0)

          ( x - x_1) * ... * (x - x_order )         ( x- 1) * (x-2) * ... * (x - order)
    L_0 = ------------------------------------  =  -------------------------------------
          (x_0 -x_1) * .... * (x_0 - x_order)        -1 * -2 * .... * -order

    where the right hand side follows after substituting x_n = n (i.e. 1 spacing). This function returns the set of
    coefficients [ a_0, a_1, ..., a_order ].

    :param order: order of the base polynomials.
    :param base_polynomial_index: which of the base polynomials to calculate
    :return: set of polynomial coefficients [ a_0, a_1, ..., a_order ]
    """
    poly = zeros(order + 1)
    poly[0] = 1
    denominator = 1
    jj = 0
    for ii in range(0, order + 1):
        if ii == base_polynomial_index:
            continue
        jj += 1

        # calculate the denomitor by repeated multiplication
        denominator = denominator * (base_polynomial_index - ii)

        # Calculate the polynomial coeficients. We start with a constant function (a_0=1) of order 0 and multiply this
        # polynomial with the next term ( x - x_0), (x-x_1) etc.
        poly[1 : jj + 1] += -ii * poly[0:jj]

    return poly / denominator


@njit(cache=True)
def integrated_lagrange_base_polynomial_coef(order, base_polynomial_index):
    """
    Calculate the polynomial coefficents of the integrated base polynomial.

    :param order: polynomial order of the interated base_polynomial.
    :param base_polynomial_index: which of the base polynomials to calculate
    :return: set of polynomial coefficients [ a_0, a_1, ..., a_[order-1], 0 ]
    """
    poly = zeros(order + 1)
    poly[0:order] = lagrange_base_polynomial_coef(order - 1, base_polynomial_index)

    # Calculate the coefficients of the integrated polynimial.
    for ii in range(order - 1):
        poly[ii] = poly[ii] / (order - ii)
    return poly


@njit(cache=True)
def evaluate_polynomial(poly, x):
    """
    Eval a polynomial at location x.
    :param poly: polynomial coeficients [a_0, a_1, ..., a_[order+1]]
    :param x: location to evaluate the polynomial/
    :return: value of the polynomial at the location
    """
    res = 0
    order = len(poly) - 1
    for ii in range(0, order + 1):
        res += poly[ii] * x ** (order - ii)
    return res


@njit(cache=True)
def integration_stencil(order: int, number_of_implicit_points: int = 1) -> NDArray:
    """
    Find the Newton-Coastes like- integration stencil given the desired order and the number of "implicit" points.
    Specicially, let the position z at instance t[j-1] be known, and we wish to approximate z at time t[j], where
    t[j] - t[j-1] = dt  for all j, given the velocities w[j]. This implies we solve

            dz
           ---- = w    ->    z[j] = z[j-1] + dz     with dz = Integral[w] ~ dt * F[w]
            dt

    To solve the integral we use Newton-Coates like approximation and express w(t) as a function of points w[j+i],
    where i = -m-1 ... n-1 using a Lagrange Polynomial. Specifically we consider points in the past and future as we
    anticipate we can buffer w values in any application.

    In this framework the interval of interest lies between j-1, and j  (i=0 and 1).

        j-m-1  ...  j-2  j-1   j   j+1  ...  j+n-1
          |    |    |    |----|    |    |    |

    The number of points used will be refered to ast the order = n+m+1. The number of points with i>=0 will be referred to as
    the number of implicit points, so that n = number_of_implicit_points. The number of points i<0 is the number of
    explicit points m = order - n - 1.

    This function calculates the weights such that

    dz  =    weights[0] w[j-m] + ... +  weights[m-1] w[j-1] + weights[m] w[j] + ... weights[order-1] w[j+n-1]

    :param order: Order of the returned Newton-Coates set of coefficients.
    :param number_of_implicit_points: number of points for which i>0
    :return: Numpy array of length Order with the weights.
    """
    weights = zeros(order)
    number_of_explicit_points = order - number_of_implicit_points

    for ii in range(0, order):
        # Get the polynomial coefficients assocated with the ii'th lagrangian base polynomial l_ii
        base_poly = integrated_lagrange_base_polynomial_coef(order, ii)

        # Calculate the dz from the evaluation of the indeterminate integrals
        weights[ii] = evaluate_polynomial(
            base_poly, number_of_explicit_points
        ) - evaluate_polynomial(base_poly, number_of_explicit_points - 1)

    return weights



from roguewave import (
    FrequencyDirectionSpectrum,
)
from roguewave.wavespectra.operations import numba_integrate_spectral_data
from roguewave.wavephysics.balance.source_term import SourceTerm
from roguewave.wavephysics.balance.solvers import (
    numba_newton_raphson,
    numba_fixed_point_iteration,
)
from xarray import DataArray, zeros_like
from typing import Literal, Tuple
from numpy import inf, empty, isnan, nan, arctan2, sqrt, pi, cos, sin, exp, any, log
from numba import njit, types, prange
from numba.typed import Dict as NumbaDict
from numpy.typing import NDArray
from roguewave.wavetheory import inverse_intrinsic_dispersion_relation

TWindInputType = Literal["u10", "friction_velocity", "ustar"]


class WindGeneration(SourceTerm):
    name = "base"

    def __init__(self, parmaters):
        super(WindGeneration, self).__init__(parmaters)
        self._wind_source_term_function = None

    def rate(
        self,
        spectrum: FrequencyDirectionSpectrum,
        speed: DataArray,
        direction: DataArray,
        roughness_length: DataArray = None,
        wind_speed_input_type: TWindInputType = "u10",
    ) -> DataArray:

        wind = (speed.values, direction.values, wind_speed_input_type)
        wind_input = _wind_generation(
            spectrum.variance_density.values,
            wind=wind,
            depth=spectrum.depth.values,
            roughness_length=roughness_length.values,
            wind_source_term_function=self._wind_source_term_function,
            spectral_grid=self.spectral_grid(spectrum),
            parameters=self.parameters,
        )
        return DataArray(data=wind_input, dims=spectrum.dims, coords=spectrum.coords())

    def bulk_rate(
        self,
        spectrum: FrequencyDirectionSpectrum,
        speed: DataArray,
        direction: DataArray,
        roughness_length: DataArray = None,
        wind_speed_input_type: TWindInputType = "u10",
    ) -> DataArray:
        wind = (speed.values, direction.values, wind_speed_input_type)
        wind_input = _bulk_wind_generation(
            spectrum.variance_density.values,
            wind,
            spectrum.depth.values,
            roughness_length.values,
            self._wind_source_term_function,
            self.spectral_grid(spectrum),
            self.parameters,
        )
        return DataArray(
            data=wind_input,
            dims=spectrum.dims_space_time,
            coords=spectrum.coords_space_time,
        )

    def roughness(
        self,
        speed: DataArray,
        direction: DataArray,
        spectrum: FrequencyDirectionSpectrum,
        roughness_length_guess: DataArray = None,
        wind_speed_input_type: TWindInputType = "u10",
    ) -> DataArray:
        if roughness_length_guess is None:
            roughness_length_guess = zeros_like(speed) - 1

        wind = (speed.values, direction.values, wind_speed_input_type)

        rougness = _roughness_estimate(
            guess=roughness_length_guess.values,
            variance_density=spectrum.variance_density.values,
            wind=wind,
            depth=spectrum.depth.values,
            wind_source_term_function=self._wind_source_term_function,
            spectral_grid=self.spectral_grid(spectrum),
            parameters=self.parameters,
        )
        return DataArray(
            data=rougness,
            dims=spectrum.dims_space_time,
            coords=spectrum.coords_space_time,
        )


@njit(cache=True, fastmath=True)
def _wave_supported_stress_point(
    wind_input: NDArray, depth, spectral_grid, parameters
) -> Tuple[NDArray, NDArray]:
    """
    :param wind_input:
    :param depth:
    :param spectral_grid:
    :param parameters:
    :return:
    """

    number_of_frequencies, number_of_directions = wind_input.shape

    radian_frequency = spectral_grid["radian_frequency"]
    radian_direction = spectral_grid["radian_direction"]
    frequency_step = spectral_grid["frequency_step"]
    direction_step = spectral_grid["direction_step"]
    gravitational_acceleration = parameters["gravitational_acceleration"]

    common_factor = (
        parameters["gravitational_acceleration"] * parameters["water_density"]
    )
    if depth == inf:
        wavenumber = radian_frequency**2 / gravitational_acceleration
    else:
        wavenumber = inverse_intrinsic_dispersion_relation(radian_frequency, depth)

    cosine_step = cos(radian_direction) * direction_step
    sine_step = sin(radian_direction) * direction_step

    stress_east = 0.0
    stress_north = 0.0
    for frequency_index in range(number_of_frequencies):
        inverse_wave_speed = (
            wavenumber[frequency_index] / radian_frequency[frequency_index]
        ) * frequency_step[frequency_index]
        for direction_index in range(number_of_directions):
            stress_east += (
                cosine_step[direction_index]
                * wind_input[frequency_index, direction_index]
                * inverse_wave_speed
            )

            stress_north += (
                sine_step[direction_index]
                * wind_input[frequency_index, direction_index]
                * inverse_wave_speed
            )

    wave_stress_magnitude = sqrt(stress_north**2 + stress_east**2) * common_factor
    wave_stress_direction = (arctan2(stress_north, stress_east) * 180 / pi) % 360

    return wave_stress_magnitude, wave_stress_direction


@njit(cache=True)
def _charnock_relation_point(friction_velocity, parameters):
    """
    Charnock relation
    :param friction_velocity:
    :param parameters:
    :return:
    """
    roughness_length = (
        friction_velocity**2
        / parameters["gravitational_acceleration"]
        * parameters["charnock_constant"]
    )

    if roughness_length > parameters["charnock_maximum_roughness"]:
        roughness_length = parameters["charnock_maximum_roughness"]
    return roughness_length


@njit()
def _roughness_estimate_point(
    guess,
    variance_density,
    wind,
    depth,
    wind_source_term_function,
    spectral_grid,
    parameters,
):
    """

    :param guess:
    :param variance_density:
    :param wind:
    :param depth:
    :param spectral_grid:
    :param parameters:
    :return:
    """

    vonkarman_constant = parameters["vonkarman_constant"]
    if guess < 0:
        if wind[2] == "u10":
            drag_coeficient_wu = (0.8 + 0.065 * wind[0]) / 1000
            guess = 10 / exp(vonkarman_constant / sqrt(drag_coeficient_wu))

        elif wind[2] in ["ustar", "friction_velocity"]:
            guess = _charnock_relation_point(wind[0], parameters)
        else:
            raise ValueError("unknown wind forcing")

    if any(isnan(variance_density)):
        return nan

    if wind[0] == 0.0:
        return nan

    generation = empty((variance_density.shape))

    function_arguments = (
        variance_density,
        wind,
        depth,
        wind_source_term_function,
        spectral_grid,
        parameters,
        generation,
    )
    return numba_fixed_point_iteration(
        _roughness_iteration_function, guess, function_arguments, bounds=(0, inf)
    )


# ----------------------------------------------------------------------------------------------------------------------
# Apply to all spatial points
# ----------------------------------------------------------------------------------------------------------------------
@njit()
def _u10_from_bulk_rate_point(
    bulk_rate,
    variance_density,
    guess_u10,
    guess_direction,
    depth,
    spectral_grid,
    parameters,
    wind_source_term_function,
    time_derivative_spectrum,
):
    """

    :param bulk_rate:
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param guess_u10:
    :param guess_direction:
    :param depth:
    :param spectral_grid:
    :param parameters:
    :return:
    """
    wind_guess = (guess_u10, guess_direction, "u10")
    roughness_memory = [-1.0]
    args = (
        roughness_memory,
        variance_density,
        wind_guess,
        depth,
        wind_source_term_function,
        spectral_grid,
        parameters,
        bulk_rate,
        time_derivative_spectrum,
    )
    if bulk_rate == 0.0:
        return 0.0

    # We are tryomg to solve for U10; ~ accuracy of 0.01m/s is sufficient. We do not really care about relative accuracy
    # - for low winds (<1m/s) answers are really inaccurate anyway
    atol = 1.0e-2
    rtol = 1.0
    numerical_stepsize = 1e-3
    try:
        u10 = numba_newton_raphson(
            _u10_iteration_function,
            guess_u10,
            args,
            (0, inf),
            atol=atol,
            rtol=rtol,
            numerical_stepsize=numerical_stepsize,
        )
    except:
        u10 = nan
    return u10


@njit()
def _wind_generation(
    variance_density,
    wind,
    depth,
    roughness_length,
    wind_source_term_function,
    spectral_grid,
    parameters,
) -> NDArray:
    (
        number_of_points,
        number_of_frequencies,
        number_of_directions,
    ) = variance_density.shape

    generation = empty((number_of_points, number_of_frequencies, number_of_directions))
    for point_index in range(number_of_points):
        wind_at_point = (wind[0][point_index], wind[1][point_index], wind[2])
        wind_generation = wind_source_term_function(
            variance_density[point_index, :, :],
            wind_at_point,
            depth[point_index],
            roughness_length[point_index],
            spectral_grid,
            parameters,
        )
        generation[point_index, :, :] = wind_generation
    return generation


@njit(cache=True)
def _bulk_wind_generation(
    variance_density,
    wind,
    depth,
    roughness_length,
    wind_source_term_function,
    spectral_grid,
    parameters,
) -> NDArray:
    number_of_points = variance_density.shape[0]
    generation = empty((number_of_points))

    for point_index in range(number_of_points):
        wind_at_point = (wind[0][point_index], wind[1][point_index], wind[2])
        wind_generation = wind_source_term_function(
            variance_density=variance_density[point_index, :, :],
            wind=wind_at_point,
            depth=depth[point_index],
            roughness_length=roughness_length[point_index],
            spectral_grid=spectral_grid,
            parameters=parameters,
        )
        generation[point_index] = numba_integrate_spectral_data(
            wind_generation, spectral_grid
        )
    return generation


@njit()
def _roughness_estimate(
    guess,
    variance_density,
    wind,
    depth,
    wind_source_term_function,
    spectral_grid,
    parameters,
) -> NDArray:
    """
    :param guess:
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param wind: wind input tuple (magnitude, direction_degrees, type_string)
    :param depth:
    :param spectral_grid:
    :param parameters:
    :return:
    """
    number_of_points = variance_density.shape[0]
    roughness = empty((number_of_points))

    for point_index in range(number_of_points):
        wind_at_point = (wind[0][point_index], wind[1][point_index], wind[2])

        if isnan(wind[0][point_index]):
            roughness[point_index] = nan
        else:
            roughness[point_index] = _roughness_estimate_point(
                guess=guess[point_index],
                variance_density=variance_density[point_index, :, :],
                wind=wind_at_point,
                depth=depth[point_index],
                wind_source_term_function=wind_source_term_function,
                spectral_grid=spectral_grid,
                parameters=parameters,
            )

    return roughness


@njit(cache=True)
def _wave_supported_stress(
    variance_density,
    wind,
    depth,
    roughness_length,
    wind_source_term_function,
    spectral_grid,
    parameters,
) -> Tuple[NDArray, NDArray]:
    """
    Calculate the wave supported wind stress.

    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param wind: wind input tuple (magnitude, direction_degrees, type_string)
    :param depth:
    :param roughness_length: Surface roughness length in meter.
    :param spectral_grid:
    :param parameters:
    :return:
    """
    number_of_points = variance_density.shape[0]
    magnitude = empty((number_of_points))
    direction = empty((number_of_points))
    for point_index in range(number_of_points):
        wind_at_point = (wind[0][point_index], wind[1][point_index], wind[2])
        wind_generation = wind_source_term_function(
            variance_density=variance_density[point_index, :, :],
            wind=wind_at_point,
            depth=depth[point_index],
            roughness_length=roughness_length[point_index],
            spectral_grid=spectral_grid,
            parameters=parameters,
        )

        (
            magnitude[point_index],
            direction[point_index],
        ) = _wave_supported_stress_point(
            wind_generation, depth[point_index], spectral_grid, parameters
        )
    return magnitude, direction


@njit(parallel=False)
def _u10_from_bulk_rate(
    bulk_rate,
    variance_density,
    guess_u10,
    guess_direction,
    depth,
    wind_source_term_function,
    parameters,
    spectral_grid,
    time_derivative_spectrum,
    progress_bar=None,
) -> NDArray:
    """

    :param bulk_rate:
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param guess_u10:
    :param guess_direction:
    :param depth:
    :param parameters:
    :param spectral_grid:
    :param progress_bar:
    :return:
    """
    number_of_points = variance_density.shape[0]
    u10 = empty((number_of_points))
    for point_index in prange(number_of_points):
        if progress_bar is not None:
            progress_bar.update(1)

        u10[point_index] = _u10_from_bulk_rate_point(
            bulk_rate[point_index],
            variance_density[point_index, :, :],
            guess_u10[point_index],
            guess_direction[point_index],
            depth[point_index],
            spectral_grid=spectral_grid,
            parameters=parameters,
            wind_source_term_function=wind_source_term_function,
            time_derivative_spectrum=time_derivative_spectrum,
        )
    return u10


# ----------------------------------------------------------------------------------------------------------------------
# Helper functions
# ----------------------------------------------------------------------------------------------------------------------
@njit()
def _roughness_iteration_function(
    roughness_length,
    variance_density,
    wind,
    depth,
    wind_source_term_function,
    spectral_grid,
    parameters,
    work_array,
):
    """
    The surface roughness is defined implicitly. We use a fixed-point iteration step to solve for the roughness. This
    is the iteration function used in the fixed point iteration.

    :param roughness_length: Surface roughness length in meter.
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param wind: wind input tuple (magnitude, direction_degrees, type_string)
    :param depth:
    :param spectral_grid:
    :param parameters:
    :return:
    """
    vonkarman_constant = parameters["vonkarman_constant"]
    elevation = parameters["elevation"]
    if wind[2] == "u10":
        friction_velocity = (
            wind[0] * vonkarman_constant / log(elevation / roughness_length)
        )
    else:
        friction_velocity = wind[0]

    # Get the wind input source term values
    work_array = wind_source_term_function(
        variance_density,
        wind,
        depth,
        roughness_length,
        spectral_grid,
        parameters,
        work_array,
    )

    # Calculate the stress
    wave_supported_stress, _ = _wave_supported_stress_point(
        work_array, depth, spectral_grid, parameters
    )

    # Given the stress ratio, correct the Charnock roughness for the presence of waves.
    total_stress = parameters["air_density"] * friction_velocity**2
    stress_ratio = wave_supported_stress / total_stress

    if stress_ratio > 0.9:
        stress_ratio = 0.9

    charnock_roughness = _charnock_relation_point(friction_velocity, parameters)
    return charnock_roughness / sqrt(1 - stress_ratio)


# ----------------------------------------------------------------------------------------------------------------------
# Helper functions
# ----------------------------------------------------------------------------------------------------------------------
@njit(fastmath=True)
def _u10_iteration_function(
    u10,
    memory_list,
    variance_density,
    wind_guess,
    depth,
    wind_source_term_function,
    spectral_grid,
    parameters,
    bulk_rate,
    time_derivative_spectrum,
):
    """
    To find the 10 meter wind that generates a certain bulk input we need to solve the inverse function for the
    wind input term. We define a function

            F(U10) = wind_bulk_rate(u10) - specified_bulk_rate

    and use a zero finder to solve for

            F(U10) = 0.

    This function defines the function F. The zero finder is responsible for calling it with the correct trailing
    arguments.

    :param wind_forcing: Wind input U10 we want to solve for.
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param wind_guess:
    :param depth:
    :param spectral_grid:
    :param parameters:
    :param bulk_rate:
    :return:
    """
    if u10 == 0.0:
        return -bulk_rate

    wind = (u10, wind_guess[1], wind_guess[2])

    # Estimate the rougness length
    roughness_length = _roughness_estimate_point(
        memory_list[0],
        variance_density,
        wind,
        depth,
        wind_source_term_function,
        spectral_grid,
        parameters,
    )
    memory_list[0] = roughness_length

    # Calculate the wind input source term values
    generation = wind_source_term_function(
        variance_density,
        wind,
        depth,
        roughness_length,
        spectral_grid,
        parameters,
    )

    # Calculate the contribution of dEdt integrated over the spectrum. We only include spectral values in region where
    # there is a positive wind input
    bulk_time_derivative_spectrum = spectral_time_derivative_in_active_region(
        time_derivative_spectrum,
        generation,
        spectral_grid,
    )

    # Integrate the input and return the difference of the current guess with the desired bulk rate
    return (
        numba_integrate_spectral_data(generation, spectral_grid)
        - bulk_rate
        - bulk_time_derivative_spectrum
    )


@njit(cache=True)
def spectral_time_derivative_in_active_region(
    time_derivative_spectrum: NDArray,
    generation: NDArray,
    spectral_grid,
):
    frequency_step = spectral_grid["frequency_step"]
    direction_step = spectral_grid["direction_step"]
    number_of_directions = time_derivative_spectrum.shape[1]
    number_of_frequencies = time_derivative_spectrum.shape[0]

    bulk_time_derivative_spectrum = 0.0

    for frequency_index in range(number_of_frequencies):
        for direction_index in range(number_of_directions):
            if generation[frequency_index, direction_index] > 0.0:
                bulk_time_derivative_spectrum += (
                    time_derivative_spectrum[frequency_index, direction_index]
                    * direction_step[direction_index]
                    * frequency_step[frequency_index]
                )
    return bulk_time_derivative_spectrum


@njit(fastmath=True)
def _spectral_grid(radian_frequency, radian_direction, frequency_step, direction_step):
    return {
        "radian_frequency": radian_frequency,
        "radian_direction": radian_direction,
        "frequency_step": frequency_step,
        "direction_step": direction_step,
    }


def _numba_parameters(**kwargs):
    _dict = NumbaDict.empty(key_type=types.unicode_type, value_type=types.float64)
    for key in kwargs:
        _dict[key] = kwargs[key]

    return _dict



"""
Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit


"""
from datetime import datetime, timezone, timedelta
from xarray import DataArray
from typing import Union, Sequence
from numpy import datetime64, ndarray, array, timedelta64
from numpy.typing import NDArray
from numbers import Number
from pandas import Series

scalar_input_types = Union[str, float, int, datetime, datetime64]
input_types = Union[scalar_input_types, Sequence[scalar_input_types]]


def to_datetime_utc(time: input_types) -> Union[datetime, Sequence[datetime], None]:
    """
    Output datetimes are garantueed to be in the UTC timezone. For timezone naive input the timezone is assumed to be
    UTC. None as input is translated to None as output to allow for cases where time is optional. Note that the
    implementation works with heterogeneous sequences.

    :param time: Time, is either a valid scalar time type or a sequence of time types.
    :return: If the input is a sequence, the output is a sequence of datetimes, otherwise it is a scalar datetime.

    """

    if time is None:
        return None

    if isinstance(time, (DataArray, list, tuple, ndarray, Series)):
        # if this is a sequence type, recursively call to datetime on the sequence
        if isinstance(time, (DataArray, Series)):
            time = time.values

        return [to_datetime_utc(x) for x in time]

    else:
        # if this is a scalar type, do the appropriate conversion.
        if isinstance(time, datetime):
            if time.tzinfo is None:
                time = time.replace(tzinfo=timezone.utc)
            return time.astimezone(timezone.utc)

        elif isinstance(time, str):
            if time[-1] == "Z":
                # From isoformat does not parse "Z" as a valid timezone designator. This should be fixed in Python 11.
                time = time[:-1] + "+00:00"
            try:
                dt = datetime.fromisoformat(time)
                if dt.tzinfo is None:
                    dt = dt.replace(tzinfo=timezone.utc)

                return dt.astimezone(
                    timezone.utc
                )  # datetime.fromisoformat(time).astimezone(timezone.utc)
            except ValueError:
                return datetime.strptime(time, "%Y-%m-%dT%H:%M:%S.%f%z").astimezone(
                    timezone.utc
                )

        elif isinstance(time, datetime64):
            # We first cast datetime64 explicitly to seconds and then as a float to allow for fractional seconds.
            return datetime.fromtimestamp(
                datetime64(time, "s").astype("float64"), tz=timezone.utc
            ).replace(tzinfo=timezone.utc)
        elif isinstance(time, Number):
            return datetime.fromtimestamp(time, tz=timezone.utc)

        else:
            raise ValueError(f"Unknown time type: {type(time)} with value {time}")


def datetime_to_iso_time_string(time: scalar_input_types):
    if time is None:
        return None

    time = to_datetime_utc(time)
    return time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def datetime64_to_timestamp(time):
    return (time - datetime64(0, "s")) / timedelta64(1, "s")


def to_datetime64(time) -> Union[None, datetime64, NDArray[datetime64]]:
    """
    Convert time input to numpy ndarrays.
    :param time:
    :return:
    """
    if time is None:
        return None

    # Convert to a (sequence of) UTC datetime(s)
    time = to_datetime_utc(time)

    if isinstance(time, datetime):
        # If a datetime- do conversion
        return datetime64(int(time.timestamp()), "s")
    else:
        # if  a sequence, do list comprehension and return an array
        return array([datetime64(int(x.timestamp()), "s") for x in time])


def time_from_timeint(t: int) -> timedelta:
    """
    unpack a timedelta from a time given as an integer in the form "hhmmss" e.g. 201813 for 20:18:13
    """
    if t >= 10000:
        hours = t // 10000
        minutes = (t - hours * 10000) // 100
        seconds = t - hours * 10000 - minutes * 100
    elif t >= 100:
        hours = t // 100
        minutes = t - hours * 100
        seconds = 0
    else:
        hours = t
        minutes = 0
        seconds = 0

    return timedelta(seconds=(hours * 3600 + minutes * 60 + seconds))


def date_from_dateint(t: int) -> datetime:
    """
    unpack a datetime from a date given as an integer in the form "yyyymmdd" or "yymmdd" e.g. 20221109 for 2022-11-09
    or 221109 for 2022-11-09
    """

    if t > 1000000:
        years = t // 10000
        months = (t - years * 10000) // 100
        days = t - years * 10000 - months * 100
    else:
        years = t // 10000
        months = (t - years * 10000) // 100
        days = t - years * 10000 - months * 100
        years = years + 2000

    return datetime(years, months, days, tzinfo=timezone.utc)


def datetime_from_time_and_date_integers(
    date_int: int, time_int: int, as_datetime64=False
) -> Union[datetime, datetime64]:
    """
    Convert a date and time given as integed encoded in the form "yyyymmdd" and "hhmm" _or_ "hhmmss" to a datetime
    :param date_int: integer of the form yyyymmdd
    :param time_int: time of the form "hhmm" or "hhmmss"
    :return:
    """
    dt = date_from_dateint(date_int) + time_from_timeint(time_int)
    if as_datetime64:
        return to_datetime64(dt)
    else:
        return dt



from typing import Mapping, Sequence, Union, Tuple
from pandas import DataFrame, concat
from xarray import DataArray
from datetime import timedelta

from roguewave.spotterapi.spotterapi import get_spotter_data
from roguewave.interpolate.dataset import tracks_as_dataset
from roguewave.modeldata.extract import extract_from_remote_dataset
from roguewave.interpolate.dataframe import interpolate_dataframe_time
from roguewave.modeldata.timebase import TimeSlice


# =============================================================================
def colocate_model_spotter(
    variable: Union[Sequence[str], str],
    spotter_ids: Sequence[str],
    time_slice: TimeSlice,
    model_name: str,
    cache_name: str = None,
    parallel: bool = True,
    timebase: str = "model",
    slice_remotely=False,
    return_as_dataset=False,
) -> Union[
    Tuple[Mapping[str, DataFrame], Mapping[str, DataFrame]], Tuple[DataArray, DataArray]
]:
    """
    Colocate spoter output and modek data
    :param variable: name of the variable of interest. Can be a list in which
            case all variables in the list are retrieved.
    :param spotter_ids: List of spotter ids of interest
    :param time_slice: time slice of interest.
    :param model_name: model name
    :param slice_remotely: (default False) if True Skip local cache and try to
    read directly from
        remote. This tries to avoid downloading full files and instead tries
        to only grab the data it needs. Typically slower.
        Does not work for grib files.
    :param parallel: If slice_remotely=True controls whether we download in
        parallel. To note- if parallel is enabled we need to make sure the
        main script is guarded by if __name__ == '__main__'.
    :param cache_name: name of local cache. If None, default cache setup will
        be used.
    :param timebase: Can be ['native','spotter','model'] if
        - native: output for spotters and model at spotters will be generated
                  at their own valid times
        - spotter: model data will be interpolated to spotter reporting time
        - model: Spotter data will be interpolated to model time
    :param return_as_dataset: Instead of a dictornary return everything as one
        xarray dataset ("Pandas on steroids"), where spotter_id is one of the
        dimensions of the datastructure (a "column" if you will)
    :return:
    """

    if return_as_dataset:
        if (not timebase == "model") and (len(spotter_ids) > 1):
            raise ValueError(
                "Cannot return as an xarray dataset if the "
                "time base is native or spotter"
            )

    if timebase.lower() == "model":
        # If we have a model time base we interpolate spotter observations onto the model grid
        # To ensure we have data to use for interpolation at the end points we widen the interval there
        obs_delta = timedelta(hours=3)
    else:
        obs_delta = timedelta(hours=0)

    df = get_spotter_data(
        spotter_ids,
        "waves",
        time_slice.start_time - obs_delta,
        time_slice.end_time + obs_delta,
    )
    spotters = {
        key: value.drop(columns="spotter_id")
        for (key, value) in df.groupby("spotter_id")
    }

    for spotter_id in list(spotters.keys()):
        # If only one value - pop because we cannot interpolate on 1 value.
        if spotters[spotter_id].shape[0] <= 2:
            spotters.pop(spotter_id)

    # we want model data to cover the entire period. Spotter data will be returned up to the enddate. Hence we enforce
    # inclusion of the endpoint during colocation, this will ensure that whenever there is spotter data, there is
    # enclosing model data (which I would argue is the expected behaviour).
    if timebase.lower() == "spotter":
        time_slice.endpoint = True

    model = extract_from_remote_dataset(
        spotters,
        variable,
        time_slice,
        model_name,
        slice_remotely=slice_remotely,
        parallel=parallel,
        cache_name=cache_name,
    )

    for spotter_id in spotters:
        s = spotters[spotter_id]  # type: DataFrame
        m = model[spotter_id]  # type: DataFrame
        model_time = m["time"].values
        if timebase.lower() == "native":
            pass
        elif timebase.lower() == "observed" or timebase.lower() == "spotter":
            m = interpolate_dataframe_time(m, s["time"].values)
        elif timebase.lower() == "model":
            s = interpolate_dataframe_time(s, m["time"].values)
        else:
            raise ValueError(
                f"Unknown timebase {timebase}, must be one of: "
                f"native, observed, model, or spotter"
            )

        model[spotter_id] = m
        spotters[spotter_id] = s

    if return_as_dataset:
        return tracks_as_dataset(model_time, model), tracks_as_dataset(
            model_time, spotters
        )
    else:
        model = colocate_into_single_dataframe(model)
        spotters = colocate_into_single_dataframe(spotters)
        return model, spotters


def colocated_tracks_as_dataset(*args) -> Tuple[DataArray, DataArray]:
    if len(args) == 1:
        """
        for convinience if passing immediately from the track interpolation
        """
        model, spotter = args[0]
    else:
        model, spotter = args

    keys = list(model.keys())
    time = model[keys[0]].index.values
    model = tracks_as_dataset(time, model)
    spotter = tracks_as_dataset(time, spotter)
    return spotter, model


def colocate_into_single_dataframe(data: dict[str, DataFrame]) -> DataFrame:
    _ids = list(data.keys())
    values = [data[_id] for _id in _ids]

    data = concat(values, keys=_ids, names=["spotter_id", "time index"])
    data.reset_index(inplace=True)
    data.drop(columns="time index", inplace=True)
    return data



"""
Contents: Routines to get data from the spotter api

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines to calculate times we have output at different init and forecasthours

Functions:

- `get_satellite_data`, get observational data from available satellites
- `get_satellite_available`, get list of available satellites
"""


# Import
# =============================================================================
from typing import Sequence, Literal, List, Dict, Union
from datetime import datetime, timedelta
from pandas import DataFrame
from roguewave.modeldata.timebase import ModelTimeConfiguration, TimeSliceLead
import boto3
from roguewave import filecache
import xarray
import netCDF4
import pandas
import numpy
from os import remove, rename


SATELLITES = Literal["jason-3", "saral", "sentinal-6"]


satellite_formats = {
    "jason-3": {
        "prefix": "JA3",
        "mapping_variable_name_to_sofar": {
            "wind_speed_alt": "windVelocity10Meter",
            "swh_ocean": "significantWaveHeight",
            "time": "time",
            "latitude": "latitude",
            "longitude": "longitude",
        },
    },
    "sentinal-6": {
        "prefix": "S6A",
        "mapping_variable_name_to_sofar": {
            "wind_speed_alt": "windVelocity10Meter",
            "swh_ocean": "significantWaveHeight",
            "time": "time",
            "latitude": "latitude",
            "longitude": "longitude",
        },
    },
    "saral": {
        "prefix": "SRL",
        "mapping_variable_name_to_sofar": {
            "wind_speed_alt": "windVelocity10Meter",
            "swh": "significantWaveHeight",
            "time": "time",
            "lat": "latitude",
            "lon": "longitude",
        },
    },
}


def _get_from_cache(
    aws_keys, output_sampling_interval: timedelta, inv_mapping, variables
) -> pandas.DataFrame:
    """
    Get data from the cache (if available) or download otherwise.

    :param aws_keys: s3 aws keys
    :param output_sampling_interval: desired sample rate of output (if None, use native sample rate)
    :param inv_mapping: variable name mapping from name in netcdf stored remotely to Sofar name
    :param variables: which variables we want to extract from the remote files (Sofar convention)
    :return:
    """

    def postprocess(filepath: str):
        data = _open_grouped_netcdf_file(filepath, inv_mapping, variables)

        if output_sampling_interval is not None:
            data = _resample_dataframe(data, output_sampling_interval)

        dataset = xarray.Dataset.from_dataframe(data)
        dataset.to_netcdf(filepath + ".nc", engine="netcdf4")
        dataset.close()
        remove(filepath)
        rename(filepath + ".nc", filepath)
        return None

    def validate(filepath: str) -> bool:
        data = xarray.open_dataset(filepath)
        for var in variables:
            if var not in data:
                data.close()
                return False
        data.close()
        return True

    if output_sampling_interval is None:
        sampling = "None"
    else:
        sampling = str(output_sampling_interval.total_seconds())

    # Add processing for grib files
    filecache.set_directive_function("postprocess", f"satellite{sampling}", postprocess)
    filecache.set_directive_function("validate", f"satellite{sampling}", validate)
    aws_keys = [
        f"validate=satellite{sampling};postprocess=satellite{sampling}:{x}"
        for x in aws_keys
    ]

    # Add a "comment" to the uri to make it unique for the given sampling. If we change sampling
    # we need to re-download the original files.
    comment = f"<<{sampling}" + "_".join(variables)
    # print(comment)
    aws_keys = [key + comment for key in aws_keys]

    # Load data into cache and get filenames

    filepaths = filecache.filepaths(aws_keys)

    # Remove processing so that the cache is in the same state as before
    filecache.remove_directive_function("postprocess", f"satellite{sampling}")
    filecache.remove_directive_function("validate", f"satellite{sampling}")

    out = []
    for file in filepaths:
        data = xarray.open_dataset(file)
        # Check if the dataset contains any data- if not skip it. Xarray will otherwise crash in trying to open it.
        if data.dims["index"] < 1:
            data.close()
            continue

        out.append(data.to_dataframe())
        data.close()
        # out.append(xarray.open_dataset( file ).to_dataframe())

    if len(out) < 1:
        return None

    else:
        # Concatenate individual data frames into a single dataframe for the satellite
        return pandas.concat(out)


def get_satellite_data(
    start_date: datetime,
    end_date: datetime,
    satellites: Sequence[SATELLITES] = None,
    variables=("significantWaveHeight", "windVelocity10Meter"),
    output_sampling_interval=None,
) -> DataFrame:
    """
    Download satellite data that is stored on s3 for the given interval specificed by start date and end date.

    :param start_date: start date of interval
    :param end_date: end date of interval
    :param satellites: List of satellites to download data from - options are ['jason-3', 'saral', 'sentinal-6'], if
        None is provided as input (default) data from all available satellites is downloaded.
    :param variables: List of Variables to download. Currently only ('significantWaveHeight', 'windVelocity10Meter') are
        supported.
    :param output_sampling_interval: desired output sampling rate.

    :return: Pandas DataFrame, containing columns for latitude, longitude, the requested veriables and the satellite
        name.
    """

    # If no satellite list is given assume we want data from all available satellites
    if satellites is None:
        satellites = get_satellite_available()

    # Get the aws keys associated with the request.
    aws_keys = _get_s3_keys(start_date, end_date, satellites)

    output = []
    for satellite in satellites:
        satellite: satellite_formats
        format = _get_format(satellite)

        # Construct inverse mapping from sofar name to netcdf name.
        mapping = format["mapping_variable_name_to_sofar"]
        inv_mapping = {mapping[key]: key for key in mapping}

        # Open netcdf files and load data into individual Pandas Dataframes.
        data = _get_from_cache(
            aws_keys[satellite], output_sampling_interval, inv_mapping, variables
        )

        if data is None:
            continue

        # Get rid of duplicate entries due to overlap in netcdf files
        if not data.index.name == "time":
            data = data.set_index("time")

        data = data.sort_index()
        data = data[~data.index.duplicated(keep="first")]
        data = data.reset_index()

        # add the satellite name as a column.
        data["satellite"] = satellite
        output.append(data)

    # Concatenate individual dataframes for each satellite into a single Dataframe and return.
    if len(output) < 1:
        return None
    else:
        return pandas.concat(output, ignore_index=True)


def get_raw_satellite_data(
    start_date: datetime,
    end_date: datetime,
    satellites: Sequence[SATELLITES] = None,
) -> Dict[str,List[str]]:
    """
    Download satellite data that is stored on s3 for the given interval specificed by start date and end date.

    :param start_date: start date of interval
    :param end_date: end date of interval
    :param satellites: List of satellites to download data from - options are ['jason-3', 'saral', 'sentinal-6'], if
        None is provided as input (default) data from all available satellites is downloaded.
    :param variables: List of Variables to download. Currently only ('significantWaveHeight', 'windVelocity10Meter') are
        supported.
    :param output_sampling_interval: desired output sampling rate.

    :return: Pandas DataFrame, containing columns for latitude, longitude, the requested veriables and the satellite
        name.
    """

    # If no satellite list is given assume we want data from all available satellites
    if satellites is None:
        satellites = get_satellite_available()

    # Get the aws keys associated with the request.
    aws_keys = _get_s3_keys(start_date, end_date, satellites)

    output = {}
    for satellite in satellites:
        # Load data into cache and get filenames
        output[satellite] = filecache.filepaths(aws_keys[satellite])

    return output


def get_satellite_available() -> Sequence[str]:
    return ("jason-3", "saral", "sentinal-6")


def _open_grouped_netcdf_file(file, inv_mapping, variables) -> DataFrame:
    """
    Open a netcdf file and return the requested variables. The netcdf file  (may) contain groups
    (essentially nested datasets).

    :param file: path to netcdf file
    :param inv_mapping: mapping from netcdf variable name to output variable name
    :param variables: variables to return (sofar convention)
    :return: pandas dataframe with requested variables.
    """

    def _open_groups(
        data: netCDF4.Dataset, inv_mapping, variable
    ) -> Union[None, netCDF4.Variable]:
        """
        Function to recursively search netcdf files groups for the desired variables. The first instance of a variable
        that matches the requested variable name is returned. If after traversing all groups no match is found None
        is returned

        :param data: netcdf dataset (or group)
        :param inv_mapping: mapping from sofar name to netcdf name
        :param variable: variable name (sofar convention)
        :return: netcdf variable
        """
        key = inv_mapping[variable]
        if key in data.variables:
            return data.variables[key]

        else:
            groups = data.groups
            if len(groups) > 0:
                for key, group in groups.items():
                    data = _open_groups(group, inv_mapping, variable)
                    if data is not None:
                        return data

        return None

    dataframe = DataFrame()

    # Add the coordinates to the output variables.
    variables = ["time", "latitude", "longitude"] + list(variables)
    with netCDF4.Dataset(file) as dataset:
        # Add variables to dataframe
        for variable in variables:
            # Extract the variable from the netcdf dataset (recursively search groups for desired variable). The
            # satellite data can be stored in a grouped Netcdf format.
            data = _open_groups(dataset, inv_mapping, variable)
            if data is not None:
                # If the variable is time, ensure we convert to a proper time datatype.
                if variable == "time":
                    data = netCDF4.num2date(
                        data[:],
                        data.units,
                        only_use_cftime_datetimes=False,
                        only_use_python_datetimes=True,
                    )

                # Store in dataframe, the [:] ensure we return a numpy array from the netcdf variable class.
                dataframe[variable] = data[:]

    # Ensure longitudes are within [-180, 180)
    dataframe["longitude"] = (dataframe["longitude"] + 180) % 360 - 180
    dataframe.set_index("time", inplace=True)
    return dataframe


def _get_format(satellite: SATELLITES) -> Dict:
    return satellite_formats[satellite]


def _bucket() -> str:
    return "sofar-wx-data-dev-os1"


def _get_s3_keys(
    start_date: datetime, end_date: datetime, satellites: Sequence[SATELLITES]
) -> Dict[str, List[str]]:
    """
    Get s3 keys for satellites
    :param start_date: start date
    :param end_date:  end date
    :param satellites: satellites.
    :return: Returns a dictionary with satelite names as key and as values a list of aws keys to the netcdf files that
    contain the data for that satellite withing the requested range.
    """

    bucket = boto3.resource("s3").Bucket(_bucket())

    # Construct the dat/cycles we are interested in. We use the timeslice logic here as the satellite data is stored
    # on 6 hour intervals - similar to the forecast cycles.
    time_slice = TimeSliceLead(
        start_date, end_date, timedelta(hours=0), endpoint=True, exact=True
    )

    satelite_prefixes = {
        satellite: _get_format(satellite)["prefix"] for satellite in satellites
    }

    keys = {satellite: [] for satellite in satellites}

    # Unfortunately we cannot programatically construct the keys (due to naming on s3). Instead we have to enquire to
    # contents of a folder.
    for date in time_slice.time_base(ModelTimeConfiguration()):

        # build the prefix
        time = date[0] + date[1]
        date_string = time.strftime("%Y%m%d")
        hour_string = time.strftime("%H")
        prefix = f"observations/raw-files/data-assimilation/satellite-altimeter/{date_string}/{hour_string}"

        # Loop over all objects that match the prefix
        for object in bucket.objects.filter(Prefix=prefix):

            # loop over satelites - if the satellite prefix matches - add the object to the list for that satellite.
            for satellite, prefix in satelite_prefixes.items():
                if prefix in object.key:
                    keys[satellite].append(f"s3://{_bucket()}/{object.key}")
                    break
    return keys


def _resample_dataframe(dataframe: DataFrame, output_sampling_interval) -> DataFrame:
    """
    Resample a dataframe ensuring that longitude is properly accounted for. Data must be sorted by time for this to
    work.
    :param dataframe:
    :param output_sampling_interval:
    :return:
    """
    if dataframe.index.name == "time":
        df = dataframe
    else:
        df = dataframe.set_index("time")

    if "longitude" in df:
        df["longitude"] = numpy.unwrap(df["longitude"], discont=180, period=360)

    # Resample the data to desired interval and take the average value to represent the bin
    df = df.resample(output_sampling_interval).mean().reset_index()

    # We may introduce NaN's when resampling over gaps in the data- lets drop those rows.
    df = df.dropna()

    if "longitude" in df:
        df["longitude"] = (df["longitude"] + 180) % 360 - 180
    return df



from roguewave.timeseries_analysis import (
    pipeline,
    DEFAULT_SPOTTER_PIPELINE,
    DEFAULT_DISPLACEMENT_PIPELINE,
)
from roguewave.timeseries_analysis.time_integration import (
    cumulative_distance,
    complex_response,
)
from roguewave.timeseries_analysis import estimate_frequency_spectrum
from roguewave.tools.time import datetime64_to_timestamp
from pandas import DataFrame
from roguewave import FrequencySpectrum
from .read_csv_data import read_displacement, read_gps
from .parser import apply_to_group
from numpy import real, conjugate


def displacement_from_gps_doppler_velocities(
    path, pipeline_stages=None, cache_as_netcdf=False
) -> DataFrame:
    if pipeline_stages is None:
        pipeline_stages = DEFAULT_SPOTTER_PIPELINE

    doppler_velocities = read_gps(path, cache_as_netcdf=cache_as_netcdf)

    def process(data: DataFrame):
        time = data["time"].values
        x = pipeline(time, data["u"].values, pipeline_stages)
        y = pipeline(time, data["v"].values, pipeline_stages)
        z = pipeline(time, data["w"].values, pipeline_stages)

        return DataFrame(
            data={"time": time, "x": x, "y": y, "z": z, "group id": data["group id"]}
        )

    return apply_to_group(process, doppler_velocities)


def displacement_from_gps_positions(path) -> DataFrame:
    pipeline_stages = DEFAULT_DISPLACEMENT_PIPELINE

    gps_location_data = read_gps(path, postprocess=True)

    def get_cumulative_distances(data: DataFrame):
        time = data["time"].values
        x, y = cumulative_distance(data["latitude"].values, data["longitude"].values)
        z = data["z"].values

        return DataFrame(
            data={"time": time, "x": x, "y": y, "z": z, "group id": data["group id"]}
        )

    cum_distance = apply_to_group(get_cumulative_distances, gps_location_data)

    def process(data: DataFrame):
        time = data["time"].values
        x = pipeline(time, data["x"].values, pipeline_stages)
        y = pipeline(time, data["y"].values, pipeline_stages)
        z = pipeline(time, data["z"].values, pipeline_stages)

        return DataFrame(
            data={"time": time, "x": x, "y": y, "z": z, "group id": data["group id"]}
        )

    return apply_to_group(process, cum_distance)


def spectra_from_raw_gps(path, **kwargs) -> FrequencySpectrum:
    displacement_doppler = displacement_from_gps_doppler_velocities(path, **kwargs)
    displacement_location = displacement_from_gps_positions(path)

    correct_for_numerical_integration = kwargs.get("response_correction", True)
    order = kwargs.get("order", 4)
    n = kwargs.get("n", 1)

    if kwargs.get("use_u", False):
        x = displacement_doppler["x"].values
    else:
        x = displacement_location["x"].values

    if kwargs.get("use_v", False):
        y = displacement_doppler["y"].values
    else:
        y = displacement_location["y"].values

    if kwargs.get("use_w", True):
        z = displacement_doppler["z"].values
    else:
        z = displacement_location["z"].values

    time = datetime64_to_timestamp(displacement_doppler["time"].values)
    spectrum = estimate_frequency_spectrum(time, x, y, z, **kwargs)

    if correct_for_numerical_integration:
        spectrum = spotter_frequency_response_correction(spectrum, order, n)
    return spectrum


def spectra_from_displacement(path, **kwargs) -> FrequencySpectrum:
    displacement = read_displacement(path)
    time = datetime64_to_timestamp(displacement["time"].values)

    kwargs = kwargs.copy()
    correct_for_numerical_integration = kwargs.pop("response_correction", True)
    order = kwargs.pop("order", 4)
    n = kwargs.pop("n", 1)

    spectrum = estimate_frequency_spectrum(
        time,
        displacement["x"].values,
        displacement["y"].values,
        displacement["z"].values,
        **kwargs
    )
    if correct_for_numerical_integration:
        spectrum = spotter_frequency_response_correction(spectrum, order, n)
    return spectrum


def spotter_frequency_response_correction(
    spectrum: FrequencySpectrum, order=4, n=1, sampling_frequency=2.5
) -> FrequencySpectrum:
    """
    Correct for the spectral dampening/amplification caused by numerical integration of velocities.
    :param spectrum:
    :param order:
    :param n:
    :return:
    """
    amplification_factor = complex_response(
        spectrum.frequency.values / sampling_frequency, order, n
    )
    R = real(amplification_factor * conjugate(amplification_factor))
    return spectrum.multiply(1 / R, ["frequency"])



from numpy import inf, abs
from numpy.typing import ArrayLike
from numba import njit
from xarray import DataArray
from typing import TypeVar

_T = TypeVar("_T", ArrayLike, DataArray)


@njit(fastmath=True)
def numba_fixed_point_iteration(
    function,
    guess,
    args,
    bounds=(-inf, inf),
) -> _T:
    """
    Fixed point iteration on a vector function. We want to solve the parallal problem x=F(x) where x is a vector. Instead
    of looping over each problem and solving them individualy using e.g. scipy solvers, we gain some efficiency by
    evaluating F in parallel, and doing the iteration ourselves. Only worthwhile if F is the expensive part and/or x
    is large.

    :param function:
    :param guess:
    :param max_iter:
    :param atol:
    :param rtol:
    :param caller:
    :return:
    """

    iterates = [guess, guess, guess]
    max_iter = 100
    aitken_acceleration = True
    rtol = 1e-4
    atol = 1e-4
    for current_iteration in range(1, max_iter + 1):
        # Update iterate
        aitken_step = aitken_acceleration and current_iteration % 3 == 0
        if aitken_step:
            # Every third step do an aitken acceleration step- if requested
            numerator = iterates[2] - iterates[1]
            denominator = iterates[1] - iterates[0]

            if denominator != 0.0 and numerator != denominator:
                ratio = numerator / denominator
                next_iterate = iterates[2] + ratio / (1.0 - ratio) * (numerator)
            else:
                next_iterate = iterates[2]
        else:
            next_iterate = function(iterates[2], *args)

        # Bounds check
        if next_iterate < bounds[0]:
            next_iterate = (bounds[0] - iterates[2]) * 0.5 + iterates[2]

        if next_iterate > bounds[1]:
            next_iterate = (bounds[1] - iterates[2]) * 0.5 + iterates[2]

        # Roll the iterates, make the last entry the latest estimate
        iterates.append(iterates.pop(0))
        iterates[2] = next_iterate

        # Convergence check
        scale = max(abs(iterates[1]), atol)
        absolute_difference = abs(iterates[2] - iterates[1])
        relative_difference = absolute_difference / scale

        if (absolute_difference < atol) & (
            relative_difference < rtol
        ) and not aitken_step:
            break
    else:
        raise ValueError("no convergence")

    return iterates[2]


@njit()
def numba_newton_raphson(
    function,
    guess,
    function_arguments,
    hard_bounds=(-inf, inf),
    max_iterations=100,
    aitken_acceleration=True,
    atol=1e-4,
    rtol=1e-4,
    numerical_stepsize=1e-4,
    verbose=False,
):

    # The last three iterates. Newest iterate last. We initially fill all with the guess.
    iterates = [guess, guess, guess]

    # The function evaluatio at the last three iterate, last evation last. We initially fill all with 0. so that entries
    # are invalid until we get to the 3rd iterate.
    func_evals = [0.0, 0.0, 0.0]

    # Guess root bounds. Initially this may not actually bound the root, but if the guess is reasonable, it may.
    root_bounds = [guess - 0.5 * abs(guess), guess + 0.5 * abs(guess)]

    # The function evaluated at the bounds. We may get lucky and immediately bound the root.
    func_at_bounds = [
        function(root_bounds[0], *function_arguments),
        function(root_bounds[1], *function_arguments),
    ]

    # Is the root bounded?
    root_bounded = func_at_bounds[0] * func_at_bounds[1] < 0

    # Start iteration.
    for current_iteration in range(1, max_iterations):

        # Evaluate the function at the latest iterate. First roll the list...
        func_evals.append(func_evals.pop(0))
        # ... and then update the latest point.
        func_evals[2] = function(iterates[2], *function_arguments)

        # Every 3rd step we do a Aitken series acceleration step
        aitken_step = aitken_acceleration and current_iteration % 3 == 0
        if aitken_step:
            # We do an Aitkon step
            numerator = iterates[2] - iterates[1]
            denominator = iterates[1] - iterates[0]
            ratio = numerator / denominator
            next_iterate = iterates[2] + ratio / (1.0 - ratio) * (numerator)

        else:
            # We use a regular update step. This can either be: numerical estimate for Newton Raphson, Secant, or a
            # bisection step.

            # First lets find out if our itereate is within the root bounds. Initially, when the root is not yet bounded
            # we will escape the bounds as we do gradient descent with Newton.
            if iterates[2] < root_bounds[0]:
                # our iterate is smaller than the left bound - set the left bound to the iterate, but respect the overal
                # bounds...
                root_bounds[0] = iterates[2]
                func_at_bounds[0] = func_evals[2]

            elif iterates[2] > root_bounds[1]:
                root_bounds[1] = iterates[2]
                func_at_bounds[1] = func_evals[2]

            # If we are within the current bounds, does any of the sections contain the root?
            elif func_at_bounds[0] * func_evals[2] < 0:

                func_at_bounds[1] = func_evals[2]
                root_bounds[1] = iterates[2]
            elif func_at_bounds[1] * func_evals[2] < 0:
                root_bounded = True
                func_at_bounds[0] = func_evals[2]
                root_bounds[0] = iterates[2]

            # With the boundaries updated, check if we now bound the root.
            root_bounded = func_at_bounds[0] * func_at_bounds[1] < 0

            # If we bound the root we can start using secant - if it wants to escape the bounds we will just use a
            # bisection step, if it doesn't secant  is faster than bisection.
            if root_bounded and current_iteration > 1:
                # Secant estimate
                derivative = (func_evals[2] - func_evals[1]) / (
                    iterates[2] - iterates[1]
                )

            else:
                # Finite difference estimate
                derivative = (
                    function(iterates[2] + numerical_stepsize, *function_arguments)
                    - func_evals[2]
                ) / numerical_stepsize

            # We encountered a stationary point.
            if derivative == 0.0:
                # If we have the root bounded- let's do a bisection step to keep going.
                if root_bounded:
                    update = (root_bounds[1] - root_bounds[0]) / 2

                else:
                    raise ValueError("")

            else:
                update = -func_evals[2] / derivative

            next_iterate = iterates[2] + update

        # Bounds check
        if not root_bounded:
            bounds_to_check = hard_bounds
        else:
            bounds_to_check = (root_bounds[0], root_bounds[1])

        if next_iterate < bounds_to_check[0]:
            next_iterate = (bounds_to_check[0] - iterates[2]) * 0.5 + iterates[2]

        if next_iterate > bounds_to_check[1]:
            next_iterate = (bounds_to_check[1] - iterates[2]) * 0.5 + iterates[2]

        # Roll the iterates, make the last entry the latest estimate
        iterates.append(iterates.pop(0))
        iterates[2] = next_iterate

        # Convergence check
        scale = max(abs(iterates[1]), atol)
        absolute_difference = abs(iterates[2] - iterates[1])
        relative_difference = absolute_difference / scale

        if verbose:
            # In case anybody wonders - this monstrosity is needed because Numba does not currently support converting
            # floats to strings.
            print(
                "Iteration",
                current_iteration,
                "max abs. errors:",
                absolute_difference,
                "(atol:",
                atol,
                "max rel. error:",
                relative_difference,
                "(rtol: ",
                rtol,
                ")",
            )

        if (absolute_difference < atol) & (relative_difference < rtol):
            break

    else:
        raise ValueError("no convergence")

    return iterates[2]



"""
    Motivation:
    -----------
        We want to be able to grab multiple variables from the same model
        for the same time period (e.g. a forecast) and return this as a single
        dataset containing all the variables. Complications:
            1. Output is typically stored remotely as one file for one variable
               at one specific valid-time, init-time (sofar).
            2. For a subset of models all variables are stored in a single file
               at one specific valid-time, init-time (ecmf-wam)
            3. For a subset of models multiple valid_times are stored in a
               single file, one variable per file (typical for reanalysis
               products, e.g. era5).

        In all cases we want to abstract away that the files are stored
        remotely through automatic downloading of remote resources, and provide
        a simple caching mechanism that stores files locally, but ensures
        the cache remains of requested size.
"""

from roguewave import filecache
import xarray
from typing import List, Union, Iterable
from .modelinformation import _get_resource_specification
from os import remove, rename
from roguewave.tools.time import to_datetime_utc, datetime_from_time_and_date_integers
from glob import glob
from .timebase import TimeSlice
from .keygeneration import generate_uris
import numpy
import pygrib


# Main functions to interact with module
# =============================================================================
def open_remote_dataset(
    variable: Union[List[str], str],
    time_slice: TimeSlice,
    model_name: str,
    cache_name: str = None,
) -> xarray.Dataset:
    """
    Get a local dataset associated with a forecast for a given
    model.

    :param variable: name of the variable of interest
    :param time_slice: time slice of interest.
    :param model_name: model name
    :param cache_name: name of local cache. If None, default cache setup will
        be used.
    :return: xarray dataset
    """

    print(f"Get from data from {model_name} for a forecast")

    return _open_variables(
        variable,
        lambda x: generate_uris(x, time_slice, model_name),
        model_name=model_name,
        cache_name=cache_name,
    )


# Functions internal to the module.
# =============================================================================


def _open_uris_as_dataset(
    aws_keys: List[str],
    model_variables: List[str],
    single_variables_per_file: bool,
    filetype="netcdf",
    cache_name: str = None,
    concatenation_dimension="time",
) -> xarray.Dataset:
    """
    Open a set of remote resources as a single local dataset using a local
    Cache. Datasets will be concatenated along the given concatenation
    dimension.

    :param aws_keys: List of remote resource identifiers ("aws keys")
    :param filetype: filetype of the remote resource
    :param cache_name: name of local file cache
    :param concatenation_dimension: dimension to concatenate datasets along to
        a single dataset.
    :return: dataset concatenated along time dimension
    """

    # Ask for both the files and whether the file was already in cache. We use
    # the latter to do a conversion from grib to netcdf for a freshly
    # downloaded file so we can use a unified netcdf interface and because
    # Grib == Slow.
    def postprocess(filepath: str):
        _convert_grib_to_netcdf_pygrib(filepath, model_variables)

    def validate(filepath: str) -> bool:
        if not single_variables_per_file:
            ds = xarray.open_dataset(filepath, engine="netcdf4")
            for variable in model_variables:
                if variable not in ds:
                    ds.close()
                    return False
            ds.close()
        return True

    if filetype == "grib":
        # Add processing for grib files
        filecache.set_directive_function("postprocess", "grib", postprocess)
        filecache.set_directive_function("validate", "grib", validate)
        aws_keys = [f"validate=grib;postprocess=grib:{x}" for x in aws_keys]

    # Load data into cache and get filenames
    filepaths = filecache.filepaths(aws_keys, cache_name)

    if filetype == "grib":
        # Remove processing so that the cache is in the same state as before
        filecache.remove_directive_function("postprocess", "grib")
        filecache.remove_directive_function("validate", "grib")

    datasets = [
        xarray.open_dataset(file, engine="netcdf4", decode_times=False)
        for file in filepaths
    ]

    # Convert time etc to cf conventions
    datasets = [xarray.decode_cf(dataset) for dataset in datasets]
    for dataset in datasets:
        init_time = to_datetime_utc(dataset.attrs.get("init_time"))
        if init_time is not None:
            dataset["init_time"] = [numpy.datetime64(init_time)]

    # Concatenate and return resulting dataset
    return xarray.concat(datasets, dim=concatenation_dimension)


def _open_variables(
    variables,
    key_generation_function,
    model_name,
    cache_name,
    map_to_sofar_variable_names=True,
) -> xarray.Dataset:
    """
    Open the datasets corresponding to the given variables and return as a
    single variable.

    :param variables: variable names in model or Sofar naming convention
    :param key_generation_function: function that takes a variable name as
        input and generates the keys associated with the request.
    :param model_name: name of the model
    :param cache_name: name of the cache we use.
    :param map_to_sofar_variable_names: If true, variable names in the returned
        xarray dataset correspond to Sofar conventions.
    :return: Dataset
    """

    if not isinstance(variables, Iterable) or isinstance(variables, str):
        variables = [variables]

    # Get the model description.
    aws_layout = _get_resource_specification(model_name)

    # We can specify variable names by Sofar name or equivalent model specific
    # name. Here we remap the variable names to make sure they are all
    # specified as model specific names for querying.
    if aws_layout.mapping_variable_name_model_to_sofar is not None:
        model_variable_names = [aws_layout.to_model_variable_name(x) for x in variables]
    else:
        # If no mapping is available, we assume variables already correspond
        # to model names (specifically Sofar variable name conventions).
        model_variable_names = variables

    concatenation_dimension = aws_layout.to_model_variable_name("time")
    if not aws_layout.single_variable_per_file or len(model_variable_names) == 1:
        # For a single variable there is no need to merge datasets. Further,
        # if all variables per valid_time are stored in a the same file
        # (e.g. ECMWF Grib) then we only need to load data once. To note,
        # in that case the key_generation_function will not use the variable
        # name.
        aws_keys = key_generation_function(model_variable_names[0])
        dataset = _open_uris_as_dataset(
            aws_keys=aws_keys,
            filetype=aws_layout.filetype,
            single_variables_per_file=aws_layout.single_variable_per_file,
            cache_name=cache_name,
            concatenation_dimension=concatenation_dimension,
            model_variables=model_variable_names,
        )
    else:
        datasets = []
        # If multiple variables are present, and the variables are remotely
        # stored in unique keys, we construct individual datasets for each
        # variable and merge the dataset at the end.
        for variable in model_variable_names:
            aws_keys = key_generation_function(variable)
            datasets.append(
                _open_uris_as_dataset(
                    aws_keys=aws_keys,
                    filetype=aws_layout.filetype,
                    single_variables_per_file=aws_layout.single_variable_per_file,
                    cache_name=cache_name,
                    concatenation_dimension=concatenation_dimension,
                    model_variables=model_variable_names,
                )
            )
        dataset = xarray.merge(datasets)

    # Only return requested variables. (for Sofar variables this removes the
    # mask - or MAPSTA - variable from the NetCDF. Since that information is
    # already contained as an exception value in the actual variables this
    # information is superfluous.
    dataset = dataset[model_variable_names]

    # Remap to Sofar variable naming conventions if requested (and needed).
    if (
        aws_layout.mapping_variable_name_model_to_sofar is not None
        and map_to_sofar_variable_names
    ):
        # we may not have downloaded all variables, make sure we only rename
        # those available in the dataset.
        _map = {
            key: value
            for key, value in aws_layout.mapping_variable_name_model_to_sofar.items()
            if (key in dataset)
        }
        dataset = dataset.rename(_map)

    # return the dataset
    return dataset


# Helper Functions
# =============================================================================


def _convert_grib_to_netcdf(filepath: str, model_variables) -> None:
    """
    Convert a grib file to a netcdf file

    :param filepath: filepath of the grib file
    :return: None
    """
    # open the dataset
    dataset = xarray.open_dataset(filepath, engine="cfgrib", decode_times=False)
    dataset = dataset[model_variables]

    # convert the dataset and close
    dataset.to_netcdf(filepath + ".nc")
    dataset.close()

    # delete old file and any lingering idx files
    remove(filepath)
    if grib_idx_file := glob(filepath + "*" + ".idx"):
        remove(grib_idx_file[0])

    # rename to old file name to ensure consistency.
    rename(filepath + ".nc", filepath)
    return None


def _convert_grib_to_netcdf_pygrib(filepath: str, model_variables) -> None:
    """
    Convert a grib file to a netcdf file

    :param filepath: filepath of the grib file
    :return: None
    """

    # Open the grib file
    grib = pygrib.open(filepath)
    coords = {}
    data = {}
    for variable in model_variables:
        # lets get the correct messages - using the given short or long name.
        try:
            messages = grib.select(shortName=variable)
        except ValueError:
            messages = grib.select(name=variable)

        assert len(messages) == 1, "grib files with multiple times not supported"
        message = messages[0]

        # Decode time, valid time
        init_time = datetime_from_time_and_date_integers(
            message["dataDate"], message["dataTime"], as_datetime64=True
        )
        valid_time = datetime_from_time_and_date_integers(
            message["validityDate"], message["validityTime"], as_datetime64=True
        )

        latitude, longitude = message.latlons()
        assert numpy.all(numpy.diff(latitude[0, :]) == 0)

        if variable == model_variables[0]:
            coords = {
                "longitude": longitude[0, :],
                "latitude": latitude[:, 0],
                "validtime": numpy.array(valid_time),
                "time": numpy.array(init_time),
            }

        data[variable] = xarray.DataArray(
            name=variable,
            data=message.values.filled(numpy.nan).astype("float32"),
            dims=["latitude", "longitude"],
            coords={"longitude": longitude[0, :], "latitude": latitude[:, 0]},
        )

    dataset = xarray.Dataset(data_vars=data, coords=coords)
    dataset.to_netcdf(filepath + ".nc")
    dataset.close()

    # delete old file and any lingering idx files
    remove(filepath)
    if grib_idx_file := glob(filepath + "*" + ".idx"):
        remove(grib_idx_file[0])

    # rename to old file name to ensure consistency.
    rename(filepath + ".nc", filepath)
    return None



# -*- encoding: utf-8 -*-
import setuptools

with open("README.md", "r") as file:
    readme_contents = file.read()

setuptools.setup(
    name="roguewave",
    version="0.2.02",
    license="Apache 2 License",
    install_requires=[
        "pysofar>=0.1.13",
        "numpy",
        "netCDF4",
        "pandas",
        "scipy",
        "numba",
        "tqdm",
        "h5netcdf",
        "s3fs",
        "boto3",
        "xarray",
        "requests",
        "pygrib",
        "numba",
        "numba_progress",
    ],
    description="Python package to interact with Sofar wave data",
    long_description=readme_contents,
    long_description_content_type="text/markdown",
    author="Pieter Bart Smit",
    author_email="sofaroceangithubbot@gmail.com",
    url="https://github.com/sofarocean/roguewave.git",
    package_dir={"": "src"},
    packages=setuptools.find_packages("src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    project_urls={"Sofar Ocean Site": "https://www.sofarocean.com"},
    include_package_data=True,
    package_data={"": ["*.json"]},
)
